# encoding: utf-8
"""
peer.py

Created by Thomas Mangin on 2009-08-25.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import time
# import traceback

from exabgp.bgp.timer import ReceiveTimer
from exabgp.bgp.timer import SendTimer
from exabgp.bgp.message import Message
from exabgp.bgp.message import STATE
from exabgp.bgp.message.open.capability.capability import Capability
from exabgp.bgp.message.open.capability import REFRESH
from exabgp.bgp.message.nop import NOP
from exabgp.bgp.message.update import Update
from exabgp.bgp.message.refresh import RouteRefresh
from exabgp.bgp.message.notification import Notification
from exabgp.bgp.message.notification import Notify
from exabgp.reactor.protocol import Protocol
from exabgp.reactor.network.error import NetworkError
from exabgp.reactor.api.processes import ProcessError

from exabgp.rib.change import Change

from exabgp.configuration.environment import environment
from exabgp.logger import Logger
from exabgp.logger import FakeLogger
from exabgp.logger import LazyFormat

from exabgp.util.trace import trace

from exabgp.util.panic import no_panic
from exabgp.util.panic import footer


class ACTION (object):
	CLOSE = 0x01
	LATER = 0x02
	NOW   = 0x03


class SEND (object):
	DONE    = 0x01
	NORMAL  = 0x02
	REFRESH = 0x04


# As we can not know if this is our first start or not, this flag is used to
# always make the program act like it was recovering from a failure
# If set to FALSE, no EOR and OPEN Flags set for Restart will be set in the
# OPEN Graceful Restart Capability
FORCE_GRACEFUL = True


class Interrupted (Exception):
	pass


# =========================================================================== KA
#

class KA (object):
	def __init__ (self, log, proto):
		self._generator = self._keepalive(proto)
		self.send_timer = SendTimer(log,proto.negotiated.holdtime)

	def _keepalive (self, proto):
		need_ka   = False
		generator = None

		while True:
			# SEND KEEPALIVES
			need_ka |= self.send_timer.need_ka()

			if need_ka:
				if not generator:
					generator = proto.new_keepalive()
					need_ka = False

			if not generator:
				yield False
				continue

			try:
				# try to close the generator and raise a StopIteration in one call
				generator.next()
				generator.next()
				# still running
				yield True
			except NetworkError:
				raise Notify(4,0,'problem with network while trying to send keepalive')
			except StopIteration:
				generator = None
				yield False

	def __call__ (self):
		#  True  if we need or are trying
		#  False if we do not need to send one
		try:
			return self._generator.next()
		except StopIteration:
			raise Notify(4,0,'could not send keepalive')


# Present a File like interface to socket.socket

class Peer (object):
	def __init__ (self, neighbor, reactor):
		try:
			self.logger = Logger()
			# We only to try to connect via TCP once
			self.once = environment.settings().tcp.once
			self.bind = True if environment.settings().tcp.bind else False
		except RuntimeError:
			self.logger = FakeLogger()
			self.once = True

		self.reactor = reactor
		self.neighbor = neighbor
		# The next restart neighbor definition
		self._neighbor = None

		# The peer should restart after a stop
		self._restart = True
		# The peer was restarted (to know what kind of open to send for graceful restart)
		self._restarted = FORCE_GRACEFUL
		self._reset_skip()

		# We want to remove routes which are not in the configuration anymote afte a signal to reload
		self._reconfigure = True
		# We want to send all the known routes
		self._resend_routes = SEND.DONE
		# We have new routes for the peers
		self._have_routes = True

		# We have been asked to teardown the session with this code
		self._teardown = None

		self._ = {'in':{},'out':{}}

		self._['in']['state'] = STATE.IDLE
		self._['out']['state'] = STATE.IDLE

		# value to reset 'generator' to
		self._['in']['enabled'] = False
		self._['out']['enabled'] = None if not self.neighbor.passive else False

		# the networking code
		self._['out']['proto'] = None
		self._['in']['proto'] = None

		# the networking code
		self._['out']['code'] = self._connect
		self._['in']['code'] = self._accept

		# the generator used by the main code
		# * False, the generator for this direction is down
		# * Generator, the code to run to connect or accept the connection
		# * None, the generator must be re-created
		self._['in']['generator'] = self._['in']['enabled']
		self._['out']['generator'] = self._['out']['enabled']

	def _reset (self, direction, message='',error=''):
		self._[direction]['state'] = STATE.IDLE

		if self._restart:
			if self._[direction]['proto']:
				self._[direction]['proto'].close('%s loop, peer reset, message [%s] error[%s]' % (direction,message,str(error)))
			self._[direction]['proto'] = None
			self._[direction]['generator'] = self._[direction]['enabled']
			self._teardown = None
			self._more_skip(direction)
			self.neighbor.rib.reset()

			# If we are restarting, and the neighbor definition is different, update the neighbor
			if self._neighbor:
				self.neighbor = self._neighbor
				self._neighbor = None
		else:
			self._[direction]['generator'] = False
			self._[direction]['proto'] = None

	def _stop (self, direction, message):
		self._[direction]['generator'] = False
		self._[direction]['proto'].close('%s loop, stop, message [%s]' % (direction,message))
		self._[direction]['proto'] = None

	# connection delay

	def _reset_skip (self):
		# We are currently not skipping connection attempts
		self._skip_time = time.time()
		# when we can not connect to a peer how many time (in loop) should we back-off
		self._next_skip = 0

	def _more_skip (self, direction):
		if direction != 'out':
			return
		self._skip_time = time.time() + self._next_skip
		self._next_skip = int(1 + self._next_skip * 1.2)
		if self._next_skip > 60:
			self._next_skip = 60

	# logging

	def me (self, message):
		return "peer %s ASN %-7s %s" % (self.neighbor.peer_address,self.neighbor.peer_as,message)

	# control

	def stop (self):
		self._teardown = 3
		self._restart = False
		self._restarted = False
		self._reset_skip()
		self.neighbor.rib.uncache()

	def resend (self):
		self._resend_routes = SEND.NORMAL
		self._reset_skip()

	def send_new (self, changes=None,update=None):
		if changes:
			self.neighbor.rib.outgoing.replace(changes)
		self._have_routes = self.neighbor.flush if update is None else update

	def reestablish (self, restart_neighbor=None):
		# we want to tear down the session and re-establish it
		self._teardown = 3
		self._restart = True
		self._restarted = True
		self._resend_routes = SEND.NORMAL
		self._neighbor = restart_neighbor
		self._reset_skip()

	def reconfigure (self, restart_neighbor=None):
		# we want to update the route which were in the configuration file
		self._reconfigure = True
		self._neighbor = restart_neighbor
		self._resend_routes = SEND.NORMAL
		self._neighbor = restart_neighbor

	def teardown (self, code, restart=True):
		self._restart = restart
		self._teardown = code
		self._reset_skip()

	# sockets we must monitor

	def sockets (self):
		ios = []
		for direction in ['in','out']:
			proto = self._[direction]['proto']
			if proto and proto.connection and proto.connection.io:
				ios.append(proto.connection.io)
		return ios

	def incoming (self, connection):
		# if the other side fails, we go back to idle
		if self._['in']['proto'] not in (True,False,None):
			self.logger.network('we already have a peer at this address')
			return False

		self._['in']['proto'] = Protocol(self).accept(connection)
		# Let's make sure we do some work with this connection
		self._['in']['generator'] = None
		self._['in']['state'] = STATE.CONNECT
		return True

	def established (self):
		return self._['in']['state'] == STATE.ESTABLISHED or self._['out']['state'] == STATE.ESTABLISHED

	def detailed_link_status (self):
		state_tbl = {
			STATE.IDLE : "Idle",
			STATE.ACTIVE : "Active",
			STATE.CONNECT : "Connect",
			STATE.OPENSENT : "OpenSent",
			STATE.OPENCONFIRM : "OpenConfirm",
			STATE.ESTABLISHED : "Established" }
		return state_tbl[max(self._["in"]["state"], self._["out"]["state"])]

	def negotiated_families(self):
		if self._['out']['proto']:
			families = ["%s/%s" % (x[0], x[1]) for x in self._['out']['proto'].negotiated.families]
		else:
			families = ["%s/%s" % (x[0], x[1]) for x in self.neighbor.families()]

		if len(families) > 1:
			return "[ %s ]" % " ".join(families)
		elif len(families) == 1:
			return families[0]

		return ''

	def _accept (self):
		# we can do this as Protocol is a mutable object
		proto = self._['in']['proto']

		# send OPEN
		message = Message.CODE.NOP
		for message in proto.new_open(self._restarted):
			if ord(message.TYPE) == Message.CODE.NOP:
				yield ACTION.NOW

		proto.negotiated.sent(message)

		self._['in']['state'] = STATE.OPENSENT

		# Read OPEN
		wait = environment.settings().bgp.openwait
		opentimer = ReceiveTimer(self.me,wait,1,1,'waited for open too long, we do not like stuck in active')
		# Only yield if we have not the open, otherwise the reactor can run the other connection
		# which would be bad as we need to do the collission check without going to the other peer
		for message in proto.read_open(self.neighbor.peer_address.ip):
			opentimer.check_ka(message)
			if ord(message.TYPE) == Message.CODE.NOP:
				yield ACTION.LATER

		self._['in']['state'] = STATE.OPENCONFIRM
		proto.negotiated.received(message)
		proto.validate_open()

		if self._['out']['state'] == STATE.OPENCONFIRM:
			self.logger.network('incoming connection finds the outgoing connection is in openconfirm')
			local_id = self.neighbor.router_id.packed
			remote_id = proto.negotiated.received_open.router_id.packed

			if local_id < remote_id:
				self.logger.network('closing the outgoing connection')
				self._stop('out','collision local id < remote id')
				yield ACTION.LATER
			else:
				self.logger.network('aborting the incoming connection')
				stop = Interrupted()
				stop.direction = 'in'
				raise stop

		# Send KEEPALIVE
		for message in self._['in']['proto'].new_keepalive('OPENCONFIRM'):
			yield ACTION.NOW

		# Start keeping keepalive timer
		self.recv_timer = ReceiveTimer(self.me,proto.negotiated.holdtime,4,0)
		# Read KEEPALIVE
		for message in proto.read_keepalive():
			self.recv_timer.check_ka(message)
			yield ACTION.NOW

		self._['in']['state'] = STATE.ESTABLISHED
		# let the caller know that we were sucesfull
		yield ACTION.NOW

	def _connect (self):
		# try to establish the outgoing connection

		proto = Protocol(self)
		generator = proto.connect()

		connected = False
		try:
			while not connected:
				if self._teardown:
					raise StopIteration()
				connected = generator.next()
				# we want to come back as soon as possible
				yield ACTION.LATER
		except StopIteration:
			# Connection failed
			if not connected:
				proto.close('connection to peer failed',self._['in']['state'] != STATE.ESTABLISHED)
			# A connection arrived before we could establish !
			if not connected or self._['in']['proto']:
				stop = Interrupted()
				stop.direction = 'out'
				yield ACTION.NOW
				raise stop

		self._['out']['state'] = STATE.CONNECT
		self._['out']['proto'] = proto

		# send OPEN
		# Only yield if we have not the open, otherwise the reactor can run the other connection
		# which would be bad as we need to set the state without going to the other peer
		message = Message.CODE.NOP
		for message in proto.new_open(self._restarted):
			if ord(message.TYPE) == Message.CODE.NOP:
				yield ACTION.NOW

		proto.negotiated.sent(message)

		self._['out']['state'] = STATE.OPENSENT

		# Read OPEN
		wait = environment.settings().bgp.openwait
		opentimer = ReceiveTimer(self.me,wait,1,1,'waited for open too long, we do not like stuck in active')
		for message in self._['out']['proto'].read_open(self.neighbor.peer_address.ip):
			opentimer.check_ka(message)
			# XXX: FIXME: change the whole code to use the ord and not the chr version
			# Only yield if we have not the open, otherwise the reactor can run the other connection
			# which would be bad as we need to do the collission check
			if ord(message.TYPE) == Message.CODE.NOP:
				yield ACTION.LATER

		self._['out']['state'] = STATE.OPENCONFIRM
		proto.negotiated.received(message)
		proto.validate_open()

		if self._['in']['state'] == STATE.OPENCONFIRM:
			self.logger.network('outgoing connection finds the incoming connection is in openconfirm')
			local_id = self.neighbor.router_id.packed
			remote_id = proto.negotiated.received_open.router_id.packed

			if local_id < remote_id:
				self.logger.network('aborting the outgoing connection')
				stop = Interrupted()
				stop.direction = 'out'
				raise stop
			else:
				self.logger.network('closing the incoming connection')
				self._stop('in','collision local id < remote id')
				yield ACTION.LATER

		# Send KEEPALIVE
		for message in proto.new_keepalive('OPENCONFIRM'):
			yield ACTION.NOW

		# Start keeping keepalive timer
		self.recv_timer = ReceiveTimer(self.me,proto.negotiated.holdtime,4,0)
		# Read KEEPALIVE
		for message in self._['out']['proto'].read_keepalive():
			self.recv_timer.check_ka(message)
			yield ACTION.NOW

		self._['out']['state'] = STATE.ESTABLISHED
		# let the caller know that we were sucesfull
		yield ACTION.NOW

	def _main (self, direction):
		"""yield True if we want to come back to it asap, None if nothing urgent, and False if stopped"""
		if self._teardown:
			raise Notify(6,3)

		proto = self._[direction]['proto']

		# Announce to the process BGP is up
		self.logger.network('Connected to peer %s (%s)' % (self.neighbor.name(),direction))
		if self.neighbor.api['neighbor-changes']:
			try:
				self.reactor.processes.up(self)
			except ProcessError:
				# Can not find any better error code than 6,0 !
				# XXX: We can not restart the program so this will come back again and again - FIX
				# XXX: In the main loop we do exit on this kind of error
				raise Notify(6,0,'ExaBGP Internal error, sorry.')

		send_eor = not self.neighbor.manual_eor
		new_routes = None
		self._resend_routes = SEND.NORMAL
		send_families = []

		# Every last asm message should be re-announced on restart
		for family in self.neighbor.asm:
			if family in self.neighbor.families():
				self.neighbor.messages.appendleft(self.neighbor.asm[family])

		operational = None
		refresh = None
		command_eor = None
		number = 0
		refresh_enhanced = True if proto.negotiated.refresh == REFRESH.ENHANCED else False

		self.send_ka = KA(self.me,proto)

		while not self._teardown:
			for message in proto.read_message():
				self.recv_timer.check_ka(message)

				if self.send_ka() is not False:
					# we need and will send a keepalive
					while self.send_ka() is None:
						yield ACTION.NOW

				# Received update
				if message.TYPE == Update.TYPE:
					number += 1

					self.logger.routes(LazyFormat(self.me('<< UPDATE (%d)' % number),message.attributes,lambda _: "%s%s" % (' attributes' if _ else '',_)))

					for nlri in message.nlris:
						self.neighbor.rib.incoming.insert_received(Change(nlri,message.attributes))
						self.logger.routes(LazyFormat(self.me('<< UPDATE (%d) nlri ' % number),nlri,str))

				elif message.TYPE == RouteRefresh.TYPE:
					if message.reserved == RouteRefresh.request:
						self._resend_routes = SEND.REFRESH
						send_families.append((message.afi,message.safi))

				# SEND OPERATIONAL
				if self.neighbor.operational:
					if not operational:
						new_operational = self.neighbor.messages.popleft() if self.neighbor.messages else None
						if new_operational:
							operational = proto.new_operational(new_operational,proto.negotiated)

					if operational:
						try:
							operational.next()
						except StopIteration:
							operational = None

				# SEND REFRESH
				if self.neighbor.route_refresh:
					if not refresh:
						new_refresh = self.neighbor.refresh.popleft() if self.neighbor.refresh else None
						if new_refresh:
							refresh = proto.new_refresh(new_refresh)

					if refresh:
						try:
							refresh.next()
						except StopIteration:
							refresh = None

				# Take the routes already sent to that peer and resend them
				if self._reconfigure:
					self._reconfigure = False

					# we are here following a configuration change
					if self._neighbor:
						# see what changed in the configuration
						self.neighbor.rib.outgoing.replace(self._neighbor.backup_changes,self._neighbor.changes)
						# do not keep the previous routes in memory as they are not useful anymore
						self._neighbor.backup_changes = []

					self._have_routes = True

				# Take the routes already sent to that peer and resend them
				if self._resend_routes != SEND.DONE:
					enhanced = True if refresh_enhanced and self._resend_routes == SEND.REFRESH else False
					self._resend_routes = SEND.DONE
					self.neighbor.rib.outgoing.resend(send_families,enhanced)
					self._have_routes = True
					send_families = []

				# Need to send update
				if self._have_routes and not new_routes:
					self._have_routes = False
					# XXX: in proto really. hum to think about ?
					new_routes = proto.new_update()

				if new_routes:
					try:
						count = 20
						while count:
							# This can raise a NetworkError
							new_routes.next()
							count -= 1
					except StopIteration:
						new_routes = None

				elif send_eor:
					send_eor = False
					for _ in proto.new_eors():
						yield ACTION.NOW
					self.logger.message(self.me('>> EOR(s)'))

				# SEND MANUAL KEEPALIVE (only if we have no more routes to send)
				elif not command_eor and self.neighbor.eor:
						new_eor = self.neighbor.eor.popleft()
						command_eor = proto.new_eors(new_eor.afi,new_eor.safi)

				if command_eor:
					try:
						command_eor.next()
					except StopIteration:
						command_eor = None

				if new_routes or message.TYPE != NOP.TYPE:
					yield ACTION.NOW
				elif self.neighbor.messages or operational:
					yield ACTION.NOW
				elif self.neighbor.eor or command_eor:
					yield ACTION.NOW
				else:
					yield ACTION.LATER

				# read_message will loop until new message arrives with NOP
				if self._teardown:
					break

		# If graceful restart, silent shutdown
		if self.neighbor.graceful_restart and proto.negotiated.sent_open.capabilities.announced(Capability.CODE.GRACEFUL_RESTART):
			self.logger.network('Closing the session without notification','error')
			proto.close('graceful restarted negotiated, closing without sending any notification')
			raise NetworkError('closing')

		# notify our peer of the shutdown
		raise Notify(6,self._teardown)

	def _run (self, direction):
		"""yield True if we want the reactor to give us back the hand with the same peer loop, None if we do not have any more work to do"""
		try:
			for action in self._[direction]['code']():
				yield action

			for action in self._main(direction):
				yield action

		# CONNECTION FAILURE
		except NetworkError,network:
			self._reset(direction,'closing connection',network)

			# we tried to connect once, it failed, we stop
			if self.once:
				self.logger.network('only one attempt to connect is allowed, stopping the peer')
				self.stop()
			return

		# NOTIFY THE PEER OF AN ERROR
		except Notify,notify:
			for direction in ['in','out']:
				if self._[direction]['proto']:
					try:
						generator = self._[direction]['proto'].new_notification(notify)
						try:
							maximum = 20
							while maximum:
								generator.next()
								maximum -= 1
								yield ACTION.NOW if maximum > 10 else ACTION.LATER
						except StopIteration:
							pass
					except (NetworkError,ProcessError):
						self.logger.network(self.me('NOTIFICATION NOT SENT'),'error')
					self._reset(direction,'notification sent (%d,%d)' % (notify.code,notify.subcode),notify)
				else:
					self._reset(direction)
			return

		# THE PEER NOTIFIED US OF AN ERROR
		except Notification,notification:
			self._reset(direction,'notification received (%d,%d)' % (notification.code,notification.subcode),notification)

			# we tried to connect once, it failed, we stop
			if self.once:
				self.logger.network('only one attempt to connect is allowed, stopping the peer')
				self.stop()
			return

		# RECEIVED a Message TYPE we did not expect
		except Message,message:
			self._reset(direction,'unexpected message received',message)
			return

		# PROBLEM WRITING TO OUR FORKED PROCESSES
		except ProcessError, process:
			self._reset(direction,'process problem',process)
			return

		# ....
		except Interrupted,interrupt:
			self._reset(interrupt.direction)
			return

		# UNHANDLED PROBLEMS
		except Exception,exc:
			# Those messages can not be filtered in purpose
			self.logger.raw('\n'.join([
				no_panic,
				self.me(''),
				'',
				str(type(exc)),
				str(exc),
				trace(),
				footer
			]))
			self._reset(direction)
			return
	# loop

	def run (self):
		back = ACTION.LATER if self._restart else ACTION.CLOSE

		for direction in ['in','out']:
			opposite = 'out' if direction == 'in' else 'in'

			generator = self._[direction]['generator']
			if generator:
				try:
					# This generator only stops when it raises
					r = generator.next()

					# if r is ACTION.NOW: status = 'immediately'
					# elif r is ACTION.LATER:   status = 'next second'
					# elif r is ACTION.CLOSE:   status = 'stop'
					# else: status = 'buggy'
					# self.logger.network('%s loop %11s, state is %s' % (direction,status,self._[direction]['state']),'debug')

					if r == ACTION.NOW:
						back = ACTION.NOW
					elif r == ACTION.LATER:
						back = ACTION.LATER if back != ACTION.NOW else ACTION.NOW
				except StopIteration:
					# Trying to run a closed loop, no point continuing
					self._[direction]['generator'] = self._[direction]['enabled']

			elif generator is None:
				if self._[opposite]['state'] in [STATE.OPENCONFIRM,STATE.ESTABLISHED]:
					self.logger.network('%s loop, stopping, other one is established' % direction,'debug')
					self._[direction]['generator'] = False
					continue
				if direction == 'out' and self._skip_time > time.time():
					self.logger.network('%s loop, skipping, not time yet' % direction,'debug')
					back = ACTION.LATER
					continue
				if self._restart:
					self.logger.network('%s loop, intialising' % direction,'debug')
					self._[direction]['generator'] = self._run(direction)
					back = ACTION.LATER  # make sure we go through a clean loop

		return back



# encoding: utf-8
"""
bgp.py

Created by Thomas Mangin on 2012-07-08.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import unpack


# =========================================================== RouteDistinguisher
# RFC 4364

class RouteDistinguisher (object):

	__slots__ = ['rd','_len']

	def __init__ (self, rd):
		self.rd = rd
		self._len = len(self.rd)

	def pack (self):
		return self.rd

	def __len__ (self):
		return self._len

	def _str (self):
		t,c1,c2,c3 = unpack('!HHHH',self.rd)
		if t == 0:
			rd = '%d:%d' % (c1,(c2 << 16)+c3)
		elif t == 1:
			rd = '%d.%d.%d.%d:%d' % (c1 >> 8,c1 & 0xFF,c2 >> 8,c2 & 0xFF,c3)
		elif t == 2:
			rd = '%d:%d' % ((c1 << 16) + c2,c3)
		else:
			rd = str(self.rd)
		return rd

	def json (self):
		if not self.rd:
			return ''
		return '"route-distinguisher": "%s"' % self._str()

	def __str__ (self):
		if not self.rd:
			return ''
		return ' route-distinguisher %s' % self._str()

	@classmethod
	def unpack (cls, data):
		return cls(data[:8])

RouteDistinguisher.NORD = RouteDistinguisher('')



# encoding: utf-8
"""
generic.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from exabgp.bgp.message.update.attribute.attribute import Attribute


# ============================================================= GenericAttribute
#

class GenericAttribute (Attribute):
	__slots__ = ['ID','FLAG','data','index']

	def __init__ (self, code, flag, data):
		self.ID = code
		self.FLAG = flag
		self.data = data
		self.index = ''

	def pack (self, negotiated=None):
		flag = self.FLAG
		length = len(self.data)
		if length > 0xFF:
			flag |= Attribute.Flag.EXTENDED_LENGTH
		if flag & Attribute.Flag.EXTENDED_LENGTH:
			len_value = pack('!H',length)
		else:
			len_value = chr(length)
		return "%s%s%s%s" % (chr(flag),chr(self.ID),len_value,self.data)

	def __len__ (self):
		return len(self.data)

	def __str__ (self):
		return '0x' + ''.join('%02x' % ord(_) for _ in self.data)

	@classmethod
	def unpack (cls, code, flag, data):
		return cls(code,flag,data)



# encoding: utf-8
"""
peer.py

Created by Thomas Mangin on 2013-02-26.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import unpack
from exabgp.protocol.ip import IPv4
from exabgp.protocol.ip import IPv6

# XXX: FIXME: Move this within Peer with the same API as other classes such as Attribute or Message


class PeerType (int):
	_str = {
		0: 'global',
		1: 'L3 VPN',
	}

	def __str__ (self):
		return self._str.get(self,'unknow %d' % self)


class PeerFlag (int):
	_v4v6 = 0b10000000

	def ipv4 (self):
		return not self & self._v4v6

	def ipv6 (self):
		return bool(self & self._v4v6)


class Peer (object):
	def __init__ (self, data):
		self.type = PeerType(ord(data[2]))
		self.flag = PeerFlag(ord(data[3]))
		self.distinguisher = unpack('!L',data[4:8])[0]
		self.asn = unpack('!L',data[28:32])[0]
		self.id = IPv4.unpack(data[32:36])

		if self.flag.ipv4():
			self.peer_address = IPv4.unpack(data[24:28])
		if self.flag.ipv6():
			self.peer_address = IPv6.unpack(data[12:28])

	def validate (self):
		return self.type in (0,1)



# encoding: utf-8
"""
capabilities.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

from exabgp.bgp.message.open.capability.capability import Capability
from exabgp.bgp.message.open.capability.addpath import AddPath
from exabgp.bgp.message.open.capability.asn4 import ASN4
from exabgp.bgp.message.open.capability.graceful import Graceful
from exabgp.bgp.message.open.capability.mp import MultiProtocol
from exabgp.bgp.message.open.capability.ms import MultiSession
from exabgp.bgp.message.open.capability.operational import Operational
from exabgp.bgp.message.open.capability.refresh import RouteRefresh
from exabgp.bgp.message.open.capability.refresh import EnhancedRouteRefresh
# from exabgp.bgp.message.open.capability.unknown import UnknownCapability

from exabgp.bgp.message.notification import Notify


# =================================================================== Parameter
#

class Parameter (int):
	AUTHENTIFICATION_INFORMATION = 0x01  # Depreciated
	CAPABILITIES                 = 0x02

	def __str__ (self):
		if self == 0x01:
			return "AUTHENTIFICATION INFORMATION"
		if self == 0x02:
			return "OPTIONAL"
		return 'UNKNOWN'

# =================================================================== Capabilities
# http://www.iana.org/assignments/capability-codes/

# +------------------------------+
# | Capability Code (1 octet)    |
# +------------------------------+
# | Capability Length (1 octet)  |
# +------------------------------+
# | Capability Value (variable)  |
# +------------------------------+


class Capabilities (dict):
	_ADD_PATH = [
		(AFI(AFI.ipv4),SAFI(SAFI.unicast)),
		(AFI(AFI.ipv6),SAFI(SAFI.unicast)),
		(AFI(AFI.ipv4),SAFI(SAFI.nlri_mpls)),
		(AFI(AFI.ipv4),SAFI(SAFI.mpls_vpn)),
		(AFI(AFI.ipv6),SAFI(SAFI.mpls_vpn)),
	]

	def announced (self, capability):
		return capability in self

	def __str__ (self):
		r = []
		for key in sorted(self.keys()):
			r.append(str(self[key]))
		return ', '.join(r)

	def _protocol (self, neighbor):
		families = neighbor.families()
		mp = MultiProtocol()
		mp.extend(families)
		self[Capability.CODE.MULTIPROTOCOL] = mp

	def _asn4 (self, neighbor):
		if not neighbor.asn4:
			return

		self[Capability.CODE.FOUR_BYTES_ASN] = ASN4(neighbor.local_as)

	def _addpath (self, neighbor):
		if not neighbor.add_path:
			return

		families = neighbor.families()
		ap_families = []
		for allowed in self._ADD_PATH:
			if allowed in families:
				ap_families.append(allowed)
		self[Capability.CODE.ADD_PATH] = AddPath(ap_families,neighbor.add_path)

	def _graceful (self, neighbor, restarted):
		if not neighbor.graceful_restart:
			return

		self[Capability.CODE.GRACEFUL_RESTART] = Graceful().set(
			Graceful.RESTART_STATE if restarted else 0x0,
			neighbor.graceful_restart,
			[(afi,safi,Graceful.FORWARDING_STATE) for (afi,safi) in neighbor.families()]
		)

	def _refresh (self, neighbor):
		if not neighbor.route_refresh:
			return
		self[Capability.CODE.ROUTE_REFRESH] = RouteRefresh()
		self[Capability.CODE.ENHANCED_ROUTE_REFRESH] = EnhancedRouteRefresh()

	def _operational (self, neighbor):
		if not neighbor.operational:
			return
		self[Capability.CODE.OPERATIONAL] = Operational()

	def _session (self, neighbor):
		if not neighbor.multisession:
			return
		# XXX: FIXME: should it not be the RFC version ?
		self[Capability.CODE.MULTISESSION] = MultiSession().set([Capability.CODE.MULTIPROTOCOL])

	def new (self, neighbor, restarted):
		self._protocol(neighbor)
		self._asn4(neighbor)
		self._addpath(neighbor)
		self._graceful(neighbor,restarted)
		self._refresh(neighbor)
		self._operational(neighbor)
		self._session(neighbor)  # MUST be the last key added, really !?! dict is not ordered !
		return self

	def pack (self):
		rs = []
		for k,capabilities in self.iteritems():
			for capability in capabilities.extract():
				rs.append("%s%s%s" % (chr(k),chr(len(capability)),capability))
		parameters = "".join(["%s%s%s" % (chr(2),chr(len(r)),r) for r in rs])
		return "%s%s" % (chr(len(parameters)),parameters)

	@staticmethod
	def unpack (data):
		def _key_values (name, data):
			if len(data) < 2:
				raise Notify(2,0,"Bad length for OPEN %s (<2) %s" % (name,Capability.hex(data)))
			l = ord(data[1])
			boundary = l+2
			if len(data) < boundary:
				raise Notify(2,0,"Bad length for OPEN %s (buffer underrun) %s" % (name,Capability.hex(data)))
			key = ord(data[0])
			value = data[2:boundary]
			rest = data[boundary:]
			return key,value,rest

		capabilities = Capabilities()

		option_len = ord(data[0])
		if not option_len:
			return capabilities

		data = data[1:option_len+1]
		while data:
			key,value,data = _key_values('parameter',data)
			# Paramaters must only be sent once.
			if key == Parameter.AUTHENTIFICATION_INFORMATION:
				raise Notify(2,5)

			if key == Parameter.CAPABILITIES:
				while value:
					capability,capv,value = _key_values('capability',value)
					capabilities[capability] = Capability.unpack(capability,capabilities,capv)
			else:
				raise Notify(2,0,'Unknow OPEN parameter %s' % hex(key))
		return capabilities



# encoding: utf-8
"""
address.py

Created by Thomas Mangin on 2012-07-16.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI


# ====================================================================== Address
#

class Address (object):
	__slots__ = ['afi','safi']

	def __init__ (self, afi, safi):
		self.afi = AFI(afi)
		self.safi = SAFI(safi)

	def family (self):
		return (self.afi,self.safi)

	def address (self):
		return "%s %s" % (str(self.afi),str(self.safi))

	def __str__ (self):
		return self.address()



# encoding: utf-8
"""
mp.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message.open.capability.capability import Capability

# ================================================================ MultiProtocol
#


class MultiProtocol (Capability,list):
	ID = Capability.CODE.MULTIPROTOCOL

	def __str__ (self):
		return 'Multiprotocol(' + ','.join(["%s %s" % (str(afi),str(safi)) for (afi,safi) in self]) + ')'

	def json (self):
		return '{ "name": "multiprotocol", "families": [%s ] }' % ','.join([' "%s/%s"' % (str(afi),str(safi)) for (afi,safi) in self])

	def extract (self):
		rs = []
		for v in self:
			rs.append(pack('!H',v[0]) + pack('!H',v[1]))
		return rs

	@staticmethod
	def unpack_capability (instance, data, capability=None):  # pylint: disable=W0613
		# XXX: FIXME: we should raise if we have twice the same AFI/SAFI
		afi = AFI.unpack(data[:2])
		safi = SAFI.unpack(data[3])
		instance.append((afi,safi))
		return instance



# encoding: utf-8
"""
trace.py

Created by Thomas Mangin on 2009-09-06.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import StringIO
import traceback


def trace ():
	buff = StringIO.StringIO()
	traceback.print_exc(file=buff)
	r = buff.getvalue()
	buff.close()
	return r



# encoding: utf-8
"""
attribute/__init__.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message.update.attribute.attribute import Attribute
# from exabgp.bgp.message.update.attribute.generic import GenericAttribute
from exabgp.bgp.message.update.attribute.origin import Origin
from exabgp.bgp.message.update.attribute.aspath import ASPath
from exabgp.bgp.message.update.attribute.aspath import AS4Path
from exabgp.bgp.message.update.attribute.nexthop import NextHop
from exabgp.bgp.message.update.attribute.med import MED
from exabgp.bgp.message.update.attribute.localpref import LocalPreference
from exabgp.bgp.message.update.attribute.atomicaggregate import AtomicAggregate
from exabgp.bgp.message.update.attribute.aggregator import Aggregator
from exabgp.bgp.message.update.attribute.aggregator import Aggregator4
from exabgp.bgp.message.update.attribute.community.communities import Communities
from exabgp.bgp.message.update.attribute.community.extended.communities import ExtendedCommunities
from exabgp.bgp.message.update.attribute.originatorid import OriginatorID
from exabgp.bgp.message.update.attribute.clusterlist import ClusterList
from exabgp.bgp.message.update.attribute.mprnlri import MPRNLRI
from exabgp.bgp.message.update.attribute.mpurnlri import MPURNLRI
from exabgp.bgp.message.update.attribute.pmsi import PMSI
from exabgp.bgp.message.update.attribute.aigp import AIGP


Attribute.register_attribute(Origin)
Attribute.register_attribute(ASPath)
Attribute.register_attribute(AS4Path)
Attribute.register_attribute(NextHop)
Attribute.register_attribute(MED)
Attribute.register_attribute(LocalPreference)
Attribute.register_attribute(AtomicAggregate)
Attribute.register_attribute(Aggregator)
Attribute.register_attribute(Aggregator4)
Attribute.register_attribute(Communities)
Attribute.register_attribute(ExtendedCommunities)
Attribute.register_attribute(OriginatorID)
Attribute.register_attribute(ClusterList)
Attribute.register_attribute(MPRNLRI)
Attribute.register_attribute(MPURNLRI)
Attribute.register_attribute(PMSI)
Attribute.register_attribute(AIGP)



# encoding: utf-8
"""
show/__init__.py

Created by Thomas Mangin on 2015-01-15.
Copyright (c) 2015-2015 Exa Networks. All rights reserved.
"""

from exabgp.configuration.engine.raised import Raised
from exabgp.configuration.engine.section import Section

from exabgp.configuration.environment import environment
from exabgp.version import version

# ================================================================== show_syntax

syntax_show = """\
show {
	version
}
"""


# ================================================================ RaisedProcess

class RaisedShow (Raised):
	syntax = syntax_show


# ================================================================== SectionShow
#

class SectionShow (Section):
	syntax = syntax_show
	name = 'show'

	def enter_show (self, tokeniser):
		Section.enter_anonymous(self,tokeniser)

	def exit_show (self, tokeniser):
		pass

	def version (self, tokeniser):
		self.content['version'] = '%s %s' % (environment.application,version)
		#self.content['version'] = '%s %s' % (environment.application,environment.version)

	@classmethod
	def register (cls, registry, location):
		registry.register_class(cls)

		registry.register_hook(cls,'enter',location,'enter_show')
		registry.register_hook(cls,'action',location+['version'],'version')
		registry.register_hook(cls,'exit', location,'exit_show')



# encoding: utf-8
"""
med.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.update.attribute.attribute import Attribute


# ====================================================================== MED (4)
#

class MED (Attribute):
	ID = Attribute.CODE.MED
	FLAG = Attribute.Flag.OPTIONAL
	CACHING = True

	__slots__ = ['med','packed']

	def __init__ (self, med, packed=None):
		self.med = med
		self.packed = self._attribute(packed if packed is not None else pack('!L',med))

	def pack (self, negotiated=None):
		return self.packed

	def __len__ (self):
		return 4

	def __str__ (self):
		return str(self.med)

	def __cmp__ (self, other):
		if not isinstance(other,self.__class__):
			return -1
		if self.med != other.med:
			return -1
		return 0

	def __hash__ (self):
		return hash(self.med)

	@classmethod
	def unpack (cls, data, negotiated):
		return cls(unpack('!L',data)[0])



# encoding: utf-8
"""
community.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message.update.attribute.attribute import Attribute
from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity
from exabgp.bgp.message.update.attribute.community.communities import Communities

from exabgp.bgp.message.notification import Notify


# ===================================================== ExtendedCommunities (16)
# http://www.iana.org/assignments/bgp-extended-communities

class ExtendedCommunities (Communities):
	ID = Attribute.CODE.EXTENDED_COMMUNITY

	@staticmethod
	def unpack (data, negotiated):
		communities = ExtendedCommunities()
		while data:
			if data and len(data) < 8:
				raise Notify(3,1,'could not decode extended community %s' % str([hex(ord(_)) for _ in data]))
			communities.add(ExtendedCommunity.unpack(data[:8],negotiated))
			data = data[8:]
		return communities



# encoding: utf-8
"""
mac.py

Created by Thomas Mangin on 2012-07-08.
Copyright (c) 2014-2015 Orange. All rights reserved.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""


# ========================================================================== MAC
#

class MAC(object):

	__slots__ = ['mac','packed']

	def __init__ (self, mac=None,packed=None):
		self.mac = mac
		self.packed = packed if packed else ''.join(chr(int(_,16)) for _ in mac.split(":"))

	def __str__ (self):
		return ':'.join('%02X' % ord(_) for _ in self.packed)

	def __repr__ (self):
		return self.__str__()

	def pack (self):
		return self.packed

	# Orange code was returning 10 !
	def __len__ (self):
		return 6

	def __cmp__ (self, other):
		if not isinstance(other,self.__class__):
			return -1
		if self.packed != other.packed:
			return -1
		return 0

	# XXX: FIXME: improve for better performance ?
	def __hash__ (self):
		return hash(str(self))

	@classmethod
	def unpack (cls, data):
		return cls(':'.join('%02X' % ord(_) for _ in data[:6]),data[:6])



# encoding: utf-8
"""
od.py

Created by Thomas Mangin on 2009-09-12.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import socket


def isipv4 (address):
	try:
		socket.inet_pton(socket.AF_INET, address)
		return True
	except socket.error:
		return False


def isipv6 (address):
	try:
		socket.inet_pton(socket.AF_INET6, address)
		return True
	except socket.error:
		return False


def isip (address):
	return isipv4(address) or isipv6(address)



# encoding: utf-8
"""
change.py

Created by Thomas Mangin on 2009-11-05.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""


class Source (object):
	UNSET         = 0
	CONFIGURATION = 1
	API           = 2
	NETWORK       = 3


class Change (object):
	SOURCE = Source.UNSET

	__slots__ = ['nlri','attributes']

	def __init__ (self, nlri, attributes):
		self.nlri = nlri
		self.attributes = attributes

	def index (self):
		return '%02x%02x' % self.nlri.family() + self.nlri.index()

	def __eq__ (self, other):
		return self.nlri == other.nlri and self.attributes == other.attributes

	def __ne__ (self, other):
		return self.nlri != other.nlri or self.attributes != other.attributes

	def extensive (self):
		# If you change this you must change as well extensive in Update
		return "%s%s" % (str(self.nlri),str(self.attributes))

	def __str__ (self):
		return self.extensive()


class ConfigurationChange (Change):
	SOURCE = Source.CONFIGURATION


class APIChange (Change):
	SOURCE = Source.API


class NetworkChange (Change):
	SOURCE = Source.NETWORK



# encoding: utf-8
"""
holdtime.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import pack

# =================================================================== HoldTime


class HoldTime (int):
	def pack (self):
		return pack('!H',self)

	def keepalive (self):
		return int(self/3)

	def __len__ (self):
		return 2



# encoding: utf-8
"""
family.py

Created by Thomas Mangin on 2014-06-22.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from exabgp.configuration.engine.raised import Raised
from exabgp.configuration.engine.section import Section
from exabgp.configuration.engine.parser import boolean

from exabgp.bgp.message.open.capability.capability import Capability
from exabgp.bgp.message.open.capability.mp import MultiProtocol

from exabgp.protocol.family import known_families

from exabgp.configuration.bgp.family import SectionFamily
from exabgp.configuration.bgp.family import syntax_family


# ============================================================ syntax_capability

syntax_capability = """\
capability <name> {
	asn4 enable|disable                         # default enabled
	aigp enable|disable                         # default disabled
	operational enable|disable                  # default disabled
	multi-session enable|disable                # default disabled
	route-refresh enable|disable                # default disabled
	graceful-restart <time in second>           # default disabled
	add-path disable|send|receive|send/receive  # default disabled
	%s
}
""" % (
	'\n\t'.join((_.replace(' <name>','') for _ in syntax_family.split('\n')))
)


# ============================================================= RaisedCapability

class RaisedCapability (Raised):
	syntax = syntax_capability


# ============================================================ SectionCapability
#

class SectionCapability (Section):
	syntax = syntax_capability
	name = 'capability'

	def enter (self, tokeniser):
		Section.enter(self,tokeniser)

		self.content[Capability.CODE(Capability.CODE.FOUR_BYTES_ASN)] = True
		self.content[Capability.CODE(Capability.CODE.AIGP)] = False
		self.content[Capability.CODE(Capability.CODE.ADD_PATH)] = 0
		self.content[Capability.CODE(Capability.CODE.OPERATIONAL)] = False
		self.content[Capability.CODE(Capability.CODE.ROUTE_REFRESH)] = False
		self.content[Capability.CODE(Capability.CODE.MULTISESSION)] = False
		self.content[Capability.CODE(Capability.CODE.GRACEFUL_RESTART)] = 0

	def exit (self, tokeniser):
		if Capability.CODE(Capability.CODE.MULTIPROTOCOL) not in self.content:
			self.content[Capability.CODE(Capability.CODE.MULTIPROTOCOL)] = MultiProtocol(known_families())

	def family (self, tokeniser):
		data = self.get_section(SectionFamily.name,tokeniser)
		if data:
			self.content[Capability.CODE(Capability.CODE.MULTIPROTOCOL)] = MultiProtocol((afi,safi) for afi in sorted(data) for safi in sorted(data[afi]))
		else:
			return False

	def asn4 (self, tokeniser):
		self._check_duplicate(tokeniser,RaisedCapability)
		self.content[Capability.CODE(Capability.CODE.FOUR_BYTES_ASN)] = boolean(tokeniser,True)

	def aigp (self, tokeniser):
		self._check_duplicate(tokeniser,RaisedCapability)
		self.content[Capability.CODE(Capability.CODE.AIGP)] = boolean(tokeniser,False)

	def addpath (self, tokeniser):
		self._check_duplicate(tokeniser,RaisedCapability)
		valid_options = ('receive','send','send/receive','disable','disabled')
		ap = tokeniser()
		if ap not in valid_options:
			raise RaisedCapability(tokeniser,"%s is not a invalid add-path paramerter, options are %s" % (ap,', '.join(valid_options)))

		self.content[Capability.CODE(Capability.CODE.ADD_PATH)] = 0
		if ap.endswith('receive'):
			self.content[Capability.CODE.ADD_PATH] += 1
		if ap.startswith('send'):
			self.content[Capability.CODE.ADD_PATH] += 2

	def operational (self, tokeniser):
		self._check_duplicate(tokeniser,RaisedCapability)
		self.content[Capability.CODE(Capability.CODE.OPERATIONAL)] = boolean(tokeniser,False)

	def refresh (self, tokeniser):
		self._check_duplicate(tokeniser,RaisedCapability)
		self.content[Capability.CODE(Capability.CODE.ROUTE_REFRESH)] = boolean(tokeniser,False)

	def multisession (self, tokeniser):
		self._check_duplicate(tokeniser,RaisedCapability)
		self.content[Capability.CODE(Capability.CODE.MULTISESSION_CISCO)] = boolean(tokeniser,False)

	def graceful (self, tokeniser):
		self._check_duplicate(tokeniser,RaisedCapability)
		token = tokeniser()
		if not token.isdigit():
			raise RaisedCapability(tokeniser,"%s is not a valid option for graceful-restart, it must be a positive number smaller than 2^16" % token)

		duration = int(token)
		if duration < 0:
			raise RaisedCapability(tokeniser,"%s is not a valid option for graceful-restart, it must be a positive number smaller than 2^16" % token)
		if duration > pow(2,16):
			raise RaisedCapability(tokeniser,"%s is not a valid option for graceful-restart, it must be a positive number smaller than 2^16" % token)

		self.content[Capability.CODE(Capability.CODE.GRACEFUL_RESTART)] = duration

	@classmethod
	def register (cls, registry, location):
		registry.register_class(cls)

		# FamilySection.register(location)

		registry.register_hook(cls,'enter',location,'enter')
		registry.register_hook(cls,'exit',location,'exit')

		registry.register(SectionFamily,location+['family'])
		registry.register_hook(cls,'action',location+['family'],'family')

		registry.register_hook(cls,'action',location+['asn4'],'asn4')
		registry.register_hook(cls,'action',location+['aigp'],'aigp')
		registry.register_hook(cls,'action',location+['add-path'],'addpath')
		registry.register_hook(cls,'action',location+['operational'],'operational')
		registry.register_hook(cls,'action',location+['route-refresh'],'refresh')
		registry.register_hook(cls,'action',location+['multi-session'],'multisession')
		registry.register_hook(cls,'action',location+['graceful-restart'],'graceful')



# encoding: utf-8
"""
prefix.py

Created by Thomas Mangin on 2013-08-07.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import math
from exabgp.protocol.ip import IP


class CIDR (object):
	EOR = False
	# we can not define slots here as otherwise it conflict in Prefix
	# __slots__ = ['packed','mask','_ip']

	_mask_to_bytes = {}

	@classmethod
	def size (cls, mask):
		return cls._mask_to_bytes.get(mask,0)

	# have a .raw for the ip
	# have a .mask for the mask
	# have a .bgp with the bgp wire format of the prefix

	def __init__ (self, packed, mask):
		self.packed = packed
		self.mask = mask
		self._ip = None

	def getip (self):
		if not self._ip:
			self._ip = IP.ntop(self.packed)
		return self._ip

	ip = property(getip)

	def __str__ (self):
		return self.prefix()

	def prefix (self):
		return "%s/%s" % (self.ip,self.mask)

	def pack (self):
		return chr(self.mask) + self.packed[:CIDR.size(self.mask)]

	def packed_ip (self):
		return self.packed[:CIDR.size(self.mask)]

	# July 2014: should never be called as it is for the RIB code only
	# def index (self):
	# 	return self.pack()

	def __len__ (self):
		return CIDR.size(self.mask) + 1

	def __cmp__ (self, other):
		if not isinstance(other,self.__class__):
			return -1
		if self.packed != other.packed:
			return -1
		if self.mask != other.mask:
			return -1
		return 0

	def __hash__ (self):
		return hash(chr(self.mask)+self.packed)


for netmask in range(0,129):
	CIDR._mask_to_bytes[netmask] = int(math.ceil(float(netmask)/8))



# encoding: utf-8
"""
refresh.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message.open.capability.capability import Capability

# ================================================================= RouteRefresh
#


class RouteRefresh (Capability):
	def __init__ (self):
		self.ID = Capability.CODE.ROUTE_REFRESH

	def __str__ (self):
		if self.ID == Capability.CODE.ROUTE_REFRESH:
			return 'Route Refresh'
		return 'Cisco Route Refresh'

	def json (self):
		return '{ "name": "route-refresh", "variant": "%s" }' % ('RFC' if self.ID == Capability.CODE.ROUTE_REFRESH else 'Cisco')

	def extract (self):
		return ['']

	@staticmethod
	def unpack_capability (instance, data, capability=None):  # pylint: disable=W0613
		# XXX: FIXME: we should set that that instance was seen and raise if seen twice
		return instance

	def __eq__ (self, other):
		if not isinstance(other,RouteRefresh):
			return False
		return self.ID == other.ID


# ========================================================= EnhancedRouteRefresh
#


class EnhancedRouteRefresh (Capability):
	ID = Capability.CODE.ENHANCED_ROUTE_REFRESH

	def __str__ (self):
		return 'Enhanced Route Refresh'

	def json (self):
		return '{ "name": "enhanced-route-refresh" }'

	def extract (self):
		return ['']

	@staticmethod
	def unpack_capability (instance, data, capability=None):  # pylint: disable=W0613
		# XXX: FIXME: we should set that that instance was seen and raise if seen twice
		return instance



# encoding: utf-8
"""
dictionary.py

Created by Thomas Mangin on 2015-01-17.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from collections import defaultdict


# ===================================================================== dictdict
# an Hardcoded defaultdict with dict as method

class Dictionary (defaultdict):
	def __init__ (self):
		self.default_factory = dict



# encoding: utf-8
"""
ip/__init__.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import socket

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

# =========================================================================== IP
#


class IP (object):
	afi = None  # here for the API
	_known = dict()

	_UNICAST = SAFI(SAFI.unicast)
	_MULTICAST = SAFI(SAFI.multicast)

	_multicast_range = set(range(224,240))  # 239

	__slots__ = ['ip','packed']

	def __init__ (self):
		raise Exception("You should use IP.create() to use IP")

	def init (self, ip, packed):
		self.ip = ip
		self.packed = packed
		return self

	@staticmethod
	def pton (ip):
		return socket.inet_pton(IP.toaf(ip),ip)

	@staticmethod
	def ntop (data):
		return socket.inet_ntop(socket.AF_INET if len(data) == 4 else socket.AF_INET6,data)

	@staticmethod
	def toaf (ip):
		# the orders matters as ::FFFF:<ipv4> is an IPv6 address
		if ':' in ip:
			return socket.AF_INET6
		if '.' in ip:
			return socket.AF_INET
		raise Exception('unrecognised ip address %s' % ip)

	@staticmethod
	def toafi (ip):
		# the orders matters as ::FFFF:<ipv4> is an IPv6 address
		if ':' in ip:
			return AFI.ipv6
		if '.' in ip:
			return AFI.ipv4
		raise Exception('unrecognised ip address %s' % ip)

	@staticmethod
	def tosafi (ip):
		if ':' in ip:
			# XXX: FIXME: I assume that ::FFFF:<ip> must be treated unicast
			# if int(ip.split(':')[-1].split('.')[0]) in IP._multicast_range:
			return SAFI.unicast
		elif '.' in ip:
			if int(ip.split('.')[0]) in IP._multicast_range:
				return SAFI.multicast
			return SAFI.unicast
		raise Exception('unrecognised ip address %s' % ip)

	def ipv4 (self):
		return True if len(self.packed) == 4 else False

	def ipv6 (self):
		return False if len(self.packed) == 4 else True

	@staticmethod
	def length (afi):
		return 4 if afi == AFI.ipv4 else 16

	def index (self):
		return self.packed

	def pack (self):
		return self.packed

	def __str__ (self):
		return self.ip

	def __repr__ (self):
		return str(self)

	def __cmp__ (self, other):
		if not isinstance(other, self.__class__):
			return -1
		if self.packed == other.packed:
			return 0
		if self.packed < other.packed:
			return -1
		return 1

	def __hash__ (self):
		return hash(str(self.__class__.__name__) + self.packed)

	@classmethod
	def klass (cls, ip):
		# the orders matters as ::FFFF:<ipv4> is an IPv6 address
		if ':' in ip:
			afi = IPv6.afi
		elif '.' in ip:
			afi = IPv4.afi
		else:
			raise Exception('can not decode this ip address : %s' % ip)
		if afi in cls._known:
			return cls._known[afi]

	@classmethod
	def create (cls, ip, data=None,klass=None):
		if klass:
			return klass(ip,data)
		return cls.klass(ip)(ip,data)

	@classmethod
	def register (cls):
		cls._known[cls.afi] = cls

	@classmethod
	def unpack (cls, data, klass=None):
		return cls.create(IP.ntop(data),data,klass)


# ========================================================================= NoIP
#

class _NoIP (object):
	packed = ''

	def pack (self, data, negotiated=None):
		return ''

	def index (self):
		return ''

	def __str__ (self):
		return 'none'

NoIP = _NoIP()

# ========================================================================= IPv4
#

class IPv4 (IP):
	# lower case to match the class Address API
	afi = AFI.ipv4

	__slots__ = []

	def __init__ (self, ip, packed=None):
		self.init(ip,packed if packed else IP.pton(ip))

	def __len__ (self):
		return 4

	def unicast (self):
		return not self.multicast()

	def multicast (self):
		return ord(self.packed[0]) in set(range(224,240))  # 239 is last

	def ipv4 (self):
		return True

	def ipv6 (self):
		return False

	@staticmethod
	def pton (ip):
		return socket.inet_pton(socket.AF_INET,ip)

	@staticmethod
	def ntop (data):
		return socket.inet_ntop(socket.AF_INET,data)

	# klass is a trick for subclasses of IP/IPv4 such as NextHop / OriginatorID
	@classmethod
	def unpack (cls, data, klass=None):
		ip = socket.inet_ntop(socket.AF_INET,data)
		if klass:
			return klass(ip,data)
		return cls(ip,data)

IPv4.register()


# ========================================================================= IPv6
#

class IPv6 (IP):
	# lower case to match the class Address API
	afi = AFI.ipv6

	__slots__ = []

	def __init__ (self, ip, packed=None):
		self.init(ip,packed if packed else socket.inet_pton(socket.AF_INET6,ip))

	def __len__ (self):
		return 16

	def ipv4 (self):
		return False

	def ipv6 (self):
		return True

	def unicast (self):
		return True

	def multicast (self):
		return False

	@staticmethod
	def pton (ip):
		return socket.inet_pton(socket.AF_INET6,ip)

	@staticmethod
	def ntop (data):
		return socket.inet_ntop(socket.AF_INET6,data)

	@classmethod
	def unpack (cls, data, klass=None):
		ip6 = socket.inet_ntop(socket.AF_INET6,data)
		if klass:
			return klass(ip6)
		return cls(ip6)

IPv6.register()



# encoding: utf-8
"""
nlri.py

Created by Thomas Mangin on 2012-07-08.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import unpack
from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.protocol.ip.address import Address
from exabgp.protocol.ip import IP
from exabgp.bgp.message import IN
from exabgp.bgp.message.notification import Notify

from exabgp.bgp.message.update.nlri.cidr import CIDR

from exabgp.logger import Logger
from exabgp.logger import LazyFormat


class NLRI (Address):
	__slots__ = []

	EOR = False

	registered_nlri = dict()
	logger = None

	def __eq__ (self,other):
		return self.index() == other.index()

	def __ne__ (self,other):
		return not self.__eq__(other)

	def index (self):
		return '%s%s%s' % (self.afi,self.safi,self.pack())

	# remove this when code restructure is finished
	def pack (self, addpath=None):
		raise Exception('unimplemented')

	@staticmethod
	def register_nlri (klass, afi, safi):
		NLRI.registered_nlri['%d/%d' % (afi,safi)] = klass

	@classmethod
	def unpack (cls, afi, safi, data, addpath, nexthop, action):
		if not cls.logger:
			cls.logger = Logger()
		cls.logger.parser(LazyFormat("parsing %s/%s nlri payload " % (afi,safi),data))

		key = '%d/%d' % (afi,safi)
		if key in cls.registered_nlri:
			return cls.registered_nlri[key].unpack(afi,safi,data,addpath,nexthop,action)
		raise Notify(3,0,'trying to decode unknown family %s/%s' % (AFI(afi),SAFI(safi)))

	@staticmethod
	def _nlri (afi, safi, bgp, action, addpath):
		labels = []
		rd = ''

		if addpath:
			path_identifier = bgp[:4]
			bgp = bgp[4:]
		else:
			path_identifier = None

		mask = ord(bgp[0])
		bgp = bgp[1:]

		if SAFI(safi).has_label():
			while bgp and mask >= 8:
				label = int(unpack('!L',chr(0) + bgp[:3])[0])
				bgp = bgp[3:]
				mask -= 24  	# 3 bytes
				# The last 4 bits are the bottom of Stack
				# The last bit is set for the last label
				labels.append(label >> 4)
				# This is a route withdrawal
				if label == 0x800000 and action == IN.WITHDRAWN:
					break
				# This is a next-hop
				if label == 0x000000:
					break
				if label & 1:
					break

		if SAFI(safi).has_rd():
			mask -= 8*8  # the 8 bytes of the route distinguisher
			rd = bgp[:8]
			bgp = bgp[8:]

		if mask < 0:
			raise Notify(3,10,'invalid length in NLRI prefix')

		if not bgp and mask:
			raise Notify(3,10,'not enough data for the mask provided to decode the NLRI')

		size = CIDR.size(mask)

		if len(bgp) < size:
			raise Notify(3,10,'could not decode route with AFI %d sand SAFI %d' % (afi,safi))

		network,bgp = bgp[:size],bgp[size:]
		padding = '\0'*(IP.length(afi)-size)
		prefix = network + padding

		return labels,rd,path_identifier,mask,size,prefix,bgp



# encoding: utf-8
"""
rib/__init__.py

Created by Thomas Mangin on 2010-01-15.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.rib.store import Store


class RIB (object):

	# when we perform a configuration reload using SIGUSR, we must not use the RIB
	# without the cache, all the updates previously sent via the API are lost

	_cache = {}

	def __init__ (self, name, adjribout, families):
		self.name = name

		if name in self._cache:
			self.incoming = self._cache[name].incoming
			self.outgoing = self._cache[name].outgoing
			self.incoming.families = families
			self.outgoing.families = families
			for family in self.outgoing._seen.keys():
				if family not in families:
					del self.outgoing._seen[family]

			if adjribout:
				self.outgoing.resend(None,False)
			else:
				self.outgoing.clear()
		else:
			self.incoming = Store(families)
			self.outgoing = Store(families)
			self._cache[name] = self

		self.outgoing.cache = adjribout

	def reset (self):
		self.incoming.reset()
		self.outgoing.reset()

	def uncache(self):
		if self.name in self._cache:
			del self._cache[self.name]

	# This code was never tested ...
	def clear (self):
		self._cache[self.name].incoming = Store(self._cache[self.name].incoming.families)
		self._cache[self.name].outgoing = Store(self._cache[self.name].incoming.families)



#!/usr/bin/env python
# encoding: utf-8
"""
api.py

Created by Thomas Mangin on 2012-12-30.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import os
import socket
import time

from exabgp.bgp.message import Message
from exabgp.bgp.message import IN


class APIOptions (dict):
	def receive_parsed (self, value):
		self['receive-parsed'] = self.get('receive-parsed',False) or value

	def receive_packets (self, value):
		self['receive-packets']  = self.get('receive-packets',False) or value

	def consolidate (self, value):
		self['consolidate'] = self.get('consolidate',False) or value

	def send_packets (self, value):
		self['send-packets'] = self.get('send-packets',False) or value

	def neighbor_changes (self, value):
		self['neighbor-changes'] = self.get('neighbor-changes',False) or value

	def receive_notifications (self, value):
		self[Message.CODE.NOTIFICATION] = self.get(Message.CODE.NOTIFICATION,False) or value

	def receive_opens (self, value):
		self[Message.CODE.OPEN] = self.get(Message.CODE.OPEN,False) or value

	def receive_keepalives (self, value):
		self[Message.CODE.KEEPALIVE] = self.get(Message.CODE.KEEPALIVE,False) or value

	def receive_updates (self, value):
		self[Message.CODE.UPDATE] = self.get(Message.CODE.UPDATE,False) or value

	def receive_refresh (self, value):
		self[Message.CODE.ROUTE_REFRESH] = self.get(Message.CODE.ROUTE_REFRESH,False) or value

	def receive_operational (self, value):
		self[Message.CODE.OPERATIONAL] = self.get(Message.CODE.OPERATIONAL,False) or value

	def __missing__ (self, key):
		return False


def hexstring (value):
	def spaced (value):
		for v in value:
			yield '%02X' % ord(v)
	return ''.join(spaced(value))


class Text (object):
	def __init__ (self, version):
		self.version = version

	def _header_body (self, header, body):
		header = ' header %s' % hexstring(header) if header else ''
		body = ' body %s' % hexstring(body) if body else ''

		total_string = header+body if body else header

		return total_string

	def up (self, peer):
		return 'neighbor %s up\n' % peer.neighbor.peer_address

	def connected (self, peer):
		return 'neighbor %s connected\n' % peer.neighbor.peer_address

	def down (self, peer, reason=''):
		return 'neighbor %s down - %s\n' % (peer.neighbor.peer_address,reason)

	def shutdown (self):
		return 'shutdown %d %d\n' % (os.getpid(),os.getppid())

	def notification (self, peer, code, subcode, data):
		return 'notification code %d subcode %d data %s\n' % (code,subcode,hexstring(data))

	def receive (self, peer, category, header, body):
		return 'neighbor %s received %d header %s body %s\n' % (peer.neighbor.peer_address,category,hexstring(header),hexstring(body))

	def keepalive (self, peer, header, body):
		return 'neighbor %s keepalive%s\n' % (peer.neighbor.peer_address,self._header_body(header,body))

	def open (self, peer, direction, sent_open, header, body):
		return 'neighbor %s open direction %s version %d asn %d hold_time %s router_id %s capabilities [%s]%s\n' % (peer.neighbor.peer_address,direction,sent_open.version, sent_open.asn, sent_open.hold_time, sent_open.router_id,str(sent_open.capabilities).lower(),self._header_body(header,body))

	def send (self, peer, category, header, body):
		return 'neighbor %s sent %d header %s body %s\n' % (peer.neighbor.peer_address,category,hexstring(header),hexstring(body))

	def update (self, peer, update, header, body):
		neighbor = peer.neighbor.peer_address
		r = 'neighbor %s update start\n' % neighbor
		attributes = str(update.attributes)
		for nlri in update.nlris:
			if nlri.EOR:
				r += 'neighbor %s announced route %s\n' % (neighbor,nlri.extensive())
			elif nlri.action == IN.ANNOUNCED:  # pylint: disable=E1101
				if nlri.nexthop:
					r += 'neighbor %s announced route %s%s\n' % (neighbor,nlri.extensive(),attributes)
				else:
					# This is an EOR or Flow or ... something newer
					r += 'neighbor %s announced %s %s\n' % (neighbor,nlri.extensive(),attributes)
			else:
				r += 'neighbor %s withdrawn route %s\n' % (neighbor,nlri.extensive())
		if header or body:
			r += '%s\n' % self._header_body(header,body)
		r += 'neighbor %s update end\n' % neighbor
		return r

	def refresh (self, peer, refresh, header, body):
		header = 'header %s' % hexstring(header) if header else ''
		body = 'body %s' % hexstring(body) if body else ''
		return 'neighbor %s route-refresh afi %s safi %s %s%s\n' % (
			peer,refresh.afi,refresh.safi,refresh.reserved,self._header_body(header,body)
		)

	def _operational_advisory (self, peer, operational, header, body):
		return 'neighbor %s operational %s afi %s safi %s advisory "%s"%s' % (
			peer,operational.name,operational.afi,operational.safi,operational.data,self._header_body(header,body)
		)

	def _operational_query (self, peer, operational, header, body):
		return 'neighbor %s operational %s afi %s safi %s%s' % (
			peer,operational.name,operational.afi,operational.safi,self._header_body(header,body)
		)

	def _operational_counter (self, peer, operational, header, body):
		return 'neighbor %s operational %s afi %s safi %s router-id %s sequence %d counter %d%s' % (
			peer,operational.name,operational.afi,operational.safi,operational.routerid,operational.sequence,operational.counter,self._header_body(header,body)
		)

	def operational (self, peer, what, operational, header, body):
		if what == 'advisory':
			return self._operational_advisory(peer,operational,header,body)
		elif what == 'query':
			return self._operational_query(peer,operational,header,body)
		elif what == 'counter':
			return self._operational_counter(peer,operational,header,body)
		# elif what == 'interface':
		# 	return self._operational_interface(peer,operational)
		else:
			raise RuntimeError('the code is broken, we are trying to print a unknown type of operational message')


def nop (_): return _


class JSON (object):
	_counter = {}

	def __init__ (self, version, highres=False):
		self.version = version
		self.time = nop if highres else int

	def count (self, peer):
		address = peer.neighbor.peer_address
		number = self._counter.get(address,0) + 1
		self._counter[address] = number
		return number

	def _string (self, _):
		return '%s' % _ if issubclass(_.__class__,int) or issubclass(_.__class__,long) or ('{' in str(_)) else '"%s"' % _

	def _header (self, content, header, body, ident=None,count=None,message_type=None):
		peer     = '"host" : "%s", '   % socket.gethostname()
		pid      = '"pid" : "%s", '    % os.getpid()
		ppid     = '"ppid" : "%s", '   % os.getppid()
		counter  = '"counter": %s, '   % count if count else ''
		header   = '"header": "%s", '  % hexstring(header) if header else ''
		body     = '"body": "%s", '    % hexstring(body) if body else ''
		mtype    = '"type": "%s", '    % message_type if message_type else 'default'

		return \
			'{ '\
			'"exabgp": "%s", '\
			'"time": %s, ' \
			'%s%s%s%s%s%s%s%s' \
			'}' % (self.version,self.time(time.time()),peer,pid,ppid,counter,mtype,header,body,content)

	def _neighbor (self, peer, content):
		neighbor = peer.neighbor

		# XXX: ip: is depecated and should be removed (be careful)
		ip = '"ip": "%s", ' % neighbor.peer_address

		address = '"address": { "local": "%s", "peer": "%s"}, ' % (neighbor.local_address,neighbor.peer_address)
		asn = '"asn": { "local": "%s", "peer": "%s"}' % (neighbor.local_as,neighbor.peer_as)
		separator = ', ' if content else ''
		return \
			'"neighbor": { ' \
			'%s%s%s%s%s' \
			'} ' % (ip,address,asn,separator,content)

	def _bmp (self, neighbor, content):
		return \
			'"bmp": { ' \
			'"ip": "%s", ' \
			'%s' \
			'} ' % (neighbor,content)

	def _kv (self, extra):
		return ", ".join('"%s": %s' % (k,self._string(v)) for (k,v) in extra.iteritems())

	def _json_kv (self, extra):
		return ", ".join('"%s": %s' % (k,v.json()) for (k,v) in extra.iteritems())

	def _minimalkv (self, extra):
		return ", ".join('"%s": %s' % (k,self._string(v)) for (k,v) in extra.iteritems() if v)

	def up (self, peer):
		return self._header(self._neighbor(peer,self._kv({
			'state': 'up',
		})),'','',peer.neighbor.identificator(),self.count(peer),message_type='state')

	def connected (self, peer):
		return self._header(self._neighbor(peer,self._kv({
			'state': 'connected',
		})),'','',peer.neighbor.identificator(),self.count(peer),message_type='state')

	def down (self, peer, reason=''):
		def escape_quote (reason):
			return reason.replace('"','\\"')
		return self._header(self._neighbor(peer,self._kv({
			'state':  'down',
			'reason': escape_quote(reason),
		})),'','',peer.neighbor.identificator(),self.count(peer),message_type='state')

	def shutdown (self):
		return self._header(self._kv({
			'notification': 'shutdown',
		}),'','','',1,message_type='notification')

	def notification (self, peer, code, subcode, data):
		return self._header(self._kv({
			'notification': '{ %s } ' % self._kv({
				'code':    code,
				'subcode': subcode,
				'data':    hexstring(data),
			})
		}),'','',peer.neighbor.identificator(),self.count(peer),message_type='notification')

	def receive (self, peer, category, header, body):
		return self._header(self._neighbor(peer,self._kv({
			'message': '{ %s } ' % self._kv({
				'received': category,
				'header':   hexstring(header),
				'body':     hexstring(body),
			})
		})),'','',peer.neighbor.identificator(),self.count(peer),message_type='raw')

	def keepalive (self, peer, header, body):
		return self._header(self._neighbor(peer,''),header,body,peer.neighbor.identificator(),self.count(peer),message_type='keepalive')

	def open (self, peer, direction, sent_open, header, body):
		return self._header(self._neighbor(peer,self._kv({
			'direction': direction,
			'open':'{ %s } ' % self._kv({
				'version':      sent_open.version,
				'asn':          sent_open.asn,
				'hold_time':    sent_open.hold_time,
				'router_id':    sent_open.router_id,
				'capabilities': '{ %s }' % self._json_kv(sent_open.capabilities),
			})
		})),header,body,peer.neighbor.identificator(),self.count(peer),message_type='open')

	def send (self, peer, category, header, body):
		return self._header(self._neighbor(peer,self._kv({
			'message':'{ %s } ' % self._kv({
				'sent':   category,
				'header': hexstring(header),
				'body':   hexstring(body),
			})
		})),'','',peer.neighbor.identificator(),self.count(peer),message_type=Message.string(category))

	def _update (self, update):
		plus = {}
		minus = {}

		# all the next-hops should be the same but let's not assume it

		for nlri in update.nlris:
			nexthop = str(nlri.nexthop) if nlri.nexthop else 'null'
			if nlri.action == IN.ANNOUNCED:  # pylint: disable=E1101
				plus.setdefault(nlri.family(),{}).setdefault(nexthop,[]).append(nlri)
			if nlri.action == IN.WITHDRAWN:  # pylint: disable=E1101
				minus.setdefault(nlri.family(),[]).append(nlri)

		add = []
		for family in plus:
			s  = '"%s %s": { ' % family
			m = ''
			for nexthop in plus[family]:
				nlris = plus[family][nexthop]
				m += '"%s": { ' % nexthop
				m += ', '.join('%s' % nlri.json() for nlri in nlris)
				m += ' }, '
			s += m[:-2]
			s += ' }'
			add.append(s)

		remove = []
		for family in minus:
			nlris = minus[family]
			s  = '"%s %s": { ' % family
			s += ', '.join('%s' % nlri.json() for nlri in nlris)
			s += ' }'
			remove.append(s)

		nlri = ''

		if not add and not remove:
			if update.nlris:  # An EOR
				return update.nlris[0].json()
		if add:
			nlri += '"announce": { %s }' % ', '.join(add)
		if add and remove:
			nlri += ', '
		if remove:
			nlri += '"withdraw": { %s }' % ', '.join(remove)

		attributes = '' if not update.attributes else '"attribute": { %s }' % update.attributes.json()
		if not attributes or not nlri:
			return '"update": { %s%s } ' % (attributes,nlri)
		return '"update": { %s, %s } ' % (attributes,nlri)

	def update (self, peer, update, header, body):
		return self._header(self._neighbor(peer,self._kv({
			'message': '{ %s}' % self._update(update)})),header,body,peer.neighbor.identificator(),self.count(peer),message_type='update')

	def refresh (self, peer, refresh, header, body):
		return self._header(
			self._neighbor(
				peer,
				'"route-refresh": { "afi": "%s", "safi": "%s", "subtype": "%s" }' % (
					refresh.afi,refresh.safi,refresh.reserved
				)
			),
			header,body,peer.neighbor.identificator(),self.count(peer),message_type='refresh')

	def bmp (self, bmp, update):
		return self._header(self._bmp(bmp,self._update(update)),'','',message_type='bmp')

	def _operational_advisory (self, peer, operational, header, body):
		return self._header(
			self._neighbor(
				peer,
				'"operational": { "name": "%s", "afi": "%s", "safi": "%s", "advisory": "%s" }' % (
					operational.name,operational.afi,operational.safi,operational.data
				)
			),
			header,body,peer.neighbor.identificator(),self.count(peer),message_type='operational')

	def _operational_query (self, peer, operational, header, body):
		return self._header(
			self._neighbor(
				peer,
				'"operational": { "name": "%s", "afi": "%s", "safi": "%s" }' % (
					operational.name,operational.afi,operational.safi
				)
			),
			header,body,peer.neighbor.identificator(),self.count(peer),message_type='operational')

	def _operational_counter (self, peer, operational, header, body):
		return self._header(
			self._neighbor(
				peer,
				'"operational": { "name": "%s", "afi": "%s", "safi": "%s", "router-id": "%s", "sequence": %d, "counter": %d }' % (
					operational.name,operational.afi,operational.safi,operational.routerid,operational.sequence,operational.counter
				)
			),
			header,body,peer.neighbor.identificator(),self.count(peer),message_type='operational')

	def operational (self, peer, what, operational, header, body):
		if what == 'advisory':
			return self._operational_advisory(peer,operational,header,body)
		elif what == 'query':
			return self._operational_query(peer,operational,header,body)
		elif what == 'counter':
			return self._operational_counter(peer,operational,header,body)
		# elif what == 'interface':
		# 	return self._operational_interface(peer,operational)
		else:
			raise RuntimeError('the code is broken, we are trying to print a unknown type of operational message')



from exabgp.protocol.family import AFI
from .connection import Connection
from .tcp import create,bind
from .tcp import connect
from .tcp import MD5
from .tcp import nagle
from .tcp import TTL
from .tcp import TTLv6
from .tcp import async
from .tcp import ready
from .error import NetworkError


class Outgoing (Connection):
	direction = 'outgoing'

	def __init__ (self, afi, peer, local, port=179,md5='',ttl=None):
		Connection.__init__(self,afi,peer,local)

		self.logger.wire("Attempting connection to %s" % self.peer)

		self.peer = peer
		self.ttl = ttl
		self.afi = afi
		self.md5 = md5
		self.port = port

		try:
			self.io = create(afi)
			MD5(self.io,peer,port,md5)
			bind(self.io,local,afi)
			async(self.io,peer)
			if afi == AFI.ipv4:
				TTL(self.io,peer,ttl)
			elif afi == AFI.ipv6:
				TTLv6(self.io,peer,ttl)
			connect(self.io,peer,port,afi,md5)
			self.init = True
		except NetworkError,exc:
			self.init = False
			self.close()
			self.logger.wire("Connection failed, %s" % str(exc))

	def establish (self):
		if not self.init:
			yield False
			return

		try:
			generator = ready(self.io)
			while True:
				connected = generator.next()
				if not connected:
					yield False
					continue
				yield True
				return
		except StopIteration:
			# self.io MUST NOT be closed here, it is closed by the caller
			yield False
			return

		nagle(self.io,self.peer)
		TTL(self.io,self.peer,self.ttl)
		yield True



# encoding: utf-8
"""
tcpflags.py

Created by Thomas Mangin on 2010-02-04.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.enum import Enum

# ====================================================================== TCPFlag
# http://www.iana.org/assignments/tcp-header-flags

class TCPFlag (Enum):
	FIN    = 0x01
	SYN    = 0x02
	RST    = 0x04
	PUSH   = 0x08
	ACK    = 0x10
	URGENT = 0x20

TCPFlag.UNKNOWN = 'unknown tcp flag %s'

TCPFlag.VALUE = {
	'fin':    TCPFlag(TCPFlag.FIN),
	'syn':    TCPFlag(TCPFlag.SYN),
	'rst':    TCPFlag(TCPFlag.RST),
	'push':   TCPFlag(TCPFlag.PUSH),
	'ack':    TCPFlag(TCPFlag.ACK),
	'urgent': TCPFlag(TCPFlag.URGENT),
}

TCPFlag.NAME = {
	TCPFlag(TCPFlag.FIN):    'fin',
	TCPFlag(TCPFlag.SYN):    'syn',
	TCPFlag(TCPFlag.RST):    'rst',
	TCPFlag(TCPFlag.PUSH):   'push',
	TCPFlag(TCPFlag.ACK):    'ack',
	TCPFlag(TCPFlag.URGENT): 'urgent',
}



# encoding: utf-8
"""
bgp.py

Created by Thomas Mangin on 2012-07-08.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""


# ===================================================================== PathInfo
# RFC draft-ietf-idr-add-paths-09

class PathInfo (object):

	__slots__ = ['path_info']

	def __init__ (self, integer=None,ip=None,packed=None):
		if packed:
			self.path_info = packed
		elif ip:
			self.path_info = ''.join([chr(int(_)) for _ in ip.split('.')])
		elif integer:
			self.path_info = ''.join([chr((integer >> offset) & 0xff) for offset in [24,16,8,0]])
		else:
			self.path_info = ''
		# sum(int(a)<<offset for (a,offset) in zip(ip.split('.'), range(24, -8, -8)))

	def __len__ (self):
		return len(self.path_info)

	def json (self):
		if self.path_info:
			return '"path-information": "%s"' % '.'.join([str(ord(_)) for _ in self.path_info])
		return ''

	def __str__ (self):
		if self.path_info:
			return ' path-information %s' % '.'.join([str(ord(_)) for _ in self.path_info])
		return ''

	def pack (self):
		if self.path_info:
			return self.path_info
		return '\x00\x00\x00\x00'

PathInfo.NOPATH = PathInfo()



# encoding: utf-8
"""
encapsulation.py

Created by Thomas Mangin on 2014-06-21.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.open.asn import ASN
from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity


# ================================================================== TrafficRate
# RFC 5575

class TrafficRate (ExtendedCommunity):
	COMMUNITY_TYPE = 0x80
	COMMUNITY_SUBTYPE = 0x06

	__slots__ = ['asn','rate']

	def __init__ (self, asn, rate, community=None):
		self.asn = asn
		self.rate = rate
		ExtendedCommunity.__init__(
			self,
			community if community is not None else pack(
				"!2sHf",
				self._packedTypeSubtype(),
				asn,rate
			)
		)

	def __str__ (self):
		return "rate-limit %d" % self.rate

	@staticmethod
	def unpack (data):
		asn,rate = unpack('!Hf',data[2:8])
		return TrafficRate(ASN(asn),rate,data[:8])


# ================================================================ TrafficAction
# RFC 5575

class TrafficAction (ExtendedCommunity):
	COMMUNITY_TYPE = 0x80
	COMMUNITY_SUBTYPE = 0x07

	_sample = {
		False: 0x0,
		True:  0x2,
	}

	_terminal = {
		False: 0x0,
		True:  0x1,
	}

	__slots__ = ['sample','terminal']

	def __init__ (self, sample, terminal, community=None):
		self.sample = sample
		self.terminal = terminal
		bitmask = self._sample[sample] | self._terminal[terminal]
		ExtendedCommunity.__init__(
			self,
			community if community is not None else pack(
				'!2sLBB',
				self._packedTypeSubtype(),
				0,0,bitmask
			)
		)

	def __str__ (self):
		s = []
		if self.sample:
			s.append('sample')
		if self.terminal:
			s.append('terminal')
		return 'action %s' % '-'.join(s)

	@staticmethod
	def unpack (data):
		bit, = unpack('!B',data[7])
		sample = bool(bit & 0x02)
		terminal = bool(bit & 0x01)
		return TrafficAction(sample,terminal,data[:8])


# ============================================================== TrafficRedirect
# RFC 5575

class TrafficRedirect (ExtendedCommunity):
	COMMUNITY_TYPE = 0x80
	COMMUNITY_SUBTYPE = 0x08

	__slots__ = ['asn','target']

	def __init__ (self, asn, target, community=None):
		self.asn = asn
		self.target = target
		ExtendedCommunity.__init__(
			self,
			community if community is not None else pack(
				"!2sHL",
				self._packedTypeSubtype(),
				asn,target
			)
		)

	def __str__ (self):
		return "redirect:%s:%s" % (self.asn,self.target)

	@staticmethod
	def unpack (data):
		asn,target = unpack('!HL',data[2:8])
		return TrafficRedirect(ASN(asn),target,data[:8])


# ================================================================== TrafficMark
# RFC 5575

class TrafficMark (ExtendedCommunity):
	COMMUNITY_TYPE = 0x80
	COMMUNITY_SUBTYPE = 0x09

	__slots__ = ['dscp']

	def __init__ (self, dscp, community=None):
		self.dscp = dscp
		ExtendedCommunity.__init__(
			self,
			community if community is not None else pack(
				"!2sLBB",
				self._packedTypeSubtype(),
				0,0,dscp
			)
		)

	def __str__ (self):
		return "mark %d" % self.dscp

	@staticmethod
	def unpack (data):
		dscp, = unpack('!B',data[7])
		return TrafficMark(dscp,data[:8])


# =============================================================== TrafficNextHop
# draft-simpson-idr-flowspec-redirect-02

# XXX: FIXME: I guess this should be a subclass of NextHop or IP ..

class TrafficNextHop (ExtendedCommunity):
	COMMUNITY_TYPE = 0x08
	COMMUNITY_SUBTYPE = 0x00

	__slots__ = ['copy']

	def __init__ (self, copy, community=None):
		self.copy = copy
		ExtendedCommunity.__init__(
			self,
			community if community is not None else pack(
				"!2sLH",
				self._packedTypeSubtype(),
				0,1 if copy else 0
			)
		)

	def __str__ (self):
		return "copy-to-nexthop" if self.copy else "redirect-to-nexthop"

	@staticmethod
	def unpack (data):
		bit, = unpack('!B',data[7])
		return TrafficNextHop(bool(bit & 0x01),data[:8])


# ============================================================ TrafficRedirectIP
# RFC 5575
# If we need to provide the <IP>:<ASN> form for the FlowSpec Redirect ...

# import socket
# Do not use socket, use IPv4.ntop or pton

# TrafficRedirectASN = TrafficRedirect

# class TrafficRedirectIP (ExtendedCommunity):
# 	COMMUNITY_TYPE = 0x80
# 	COMMUNITY_SUBTYPE = 0x08

# 	def __init__ (self, ip, target, community=None):
# 		self.ip = ip
# 		self.target = target
# 		ExtendedCommunity.__init__(self,community if community is not None else pack("!BB4sH",0x80,0x08,socket.inet_pton(socket.AF_INET,ip),target))

# 	def __str__ (self):
# 		return "redirect %s:%d" % (self.ip,self.target)

# 	@staticmethod
# 	def unpack (data):
# 		ip,target = unpack('!4sH',data[2:8])
# 		return TrafficRedirectIP(socket.inet_ntop(socket.AF_INET,ip),target,data[:8])



# encoding: utf-8
"""
path.py

Created by Thomas Mangin on 2014-06-27.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.ip import IP
from exabgp.protocol.ip import NoIP
from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message.update.nlri.nlri import NLRI
from exabgp.bgp.message.update.nlri.cidr import CIDR
from exabgp.bgp.message.update.nlri.qualifier.path import PathInfo


class Prefix (CIDR,NLRI):
	__slots__ = ['path_info','nexthop','action']

	def __init__ (self, afi, safi, packed, mask, nexthop, action,path=None):
		self.path_info = PathInfo.NOPATH if path is None else path
		self.nexthop = IP.unpack(nexthop) if nexthop else NoIP
		NLRI.__init__(self,afi,safi)
		CIDR.__init__(self,packed,mask)
		self.action = action

	def prefix (self):
		return "%s/%s%s" % (CIDR.getip(self),self.mask,str(self.path_info) if self.path_info is not PathInfo.NOPATH else '')

	def extensive (self):
		return "%s/%s%s next-hop %s" % (self.ip,self.mask,str(self.path_info) if self.path_info is not PathInfo.NOPATH else '',self.nexthop)

	def pack (self, addpath):
		if addpath or True:
			return self.path_info.pack() + CIDR.pack(self) if addpath else CIDR.pack(self)
		return CIDR.pack(self) if addpath else CIDR.pack(self)

	def json (self):
		return '"%s/%s": { %s }' % (CIDR.getip(self),self.mask,self.path_info.json())

	def index (self):
		return self.pack(True)

	def __len__ (self):
		return CIDR.__len__(self) + len(self.path_info)

	def __str__ (self):
		nexthop = ' next-hop %s' % self.nexthop if self.nexthop else ''
		return "%s%s" % (self.prefix(),nexthop)

	def __hash__ (self):
		return hash(self.pack())

	@classmethod
	def unpack (cls, afi, safi, data, addpath, nexthop, action):
		labels,rd,path_identifier,mask,size,prefix,left = NLRI._nlri(afi,safi,data,action,addpath)
		nlri = cls(afi,safi,prefix,mask,nexthop,action)
		if path_identifier:
			nlri.path_info = PathInfo(None,None,path_identifier)
		return len(data) - len(left),nlri



# encoding: utf-8
"""
message.py

Created by Thomas Mangin on 2013-02-26.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import unpack
from exabgp.bmp.peer import Peer
from exabgp.bmp.message import Message


class Header (object):
	def __init__ (self, data):
		self.version = ord(data[0])
		self.message = Message(ord(data[1]))
		self.peer = Peer(data)

		self.time_sec = unpack('!L',data[36:40])[0]
		self.time_micro_sec = unpack('!L',data[40:44])[0]

	def validate (self):
		if self.version != 1:
			return False
		if not self.message.validate():
			return False
		if not self.peer.validate():
			return False
		return True

	def json (self):
		return "{}"



from exabgp.util.errstr import errstr

from .connection import Connection
from .tcp import nagle
from .tcp import async
from .error import NetworkError
from .error import NotConnected

from exabgp.bgp.message.notification import Notify


class Incoming (Connection):
	direction = 'incoming'

	def __init__ (self, afi, peer, local, io):
		Connection.__init__(self,afi,peer,local)

		self.logger.wire("Connection from %s" % self.peer)

		try:
			self.io = io
			async(self.io,peer)
			nagle(self.io,peer)
		except NetworkError,exc:
			self.close()
			raise NotConnected(errstr(exc))

	# XXX: FIXME: is that code ever called ?
	def notification (self, code, subcode, message):
		try:
			notification = Notify(code,subcode,message).message()
			for boolean in self.writer(notification):
				yield False
			# self.logger.message(self.me('>> NOTIFICATION (%d,%d,"%s")' % (notification.code,notification.subcode,notification.data)),'error')
			yield True
		except NetworkError:
			pass  # This is only be used when closing session due to unconfigured peers - so issues do not matter



# encoding: utf-8
"""
network.py

Created by Thomas Mangin on 2015-03-31.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import socket
from struct import calcsize
from collections import namedtuple

from exabgp.netlink.message import NetLinkMessage
from exabgp.netlink.message import InfoMessage
from exabgp.netlink.attributes import Attributes


# 0                   1                   2                   3
# 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |   Family    |  Src length   |  Dest length  |     TOS       |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |  Table ID   |   Protocol    |     Scope     |     Type      |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
# |                          Flags                              |
# +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

class Network (InfoMessage):
	class Header (object):
		# linux/if_addr.h
		PACK = '8BI'  # or is it 8Bi ?
		LEN = calcsize(PACK)

	format = namedtuple('Neighbor', 'family src_len dst_len tos table proto scope type flags attributes')

	class Command (object):
		RTM_NEWROUTE = 0x18
		RTM_DELROUTE = 0x19
		RTM_GETROUTE = 0x1A

	class Type (object):
		class Table (object):
			RT_TABLE_UNSPEC   = 0x00  # An unspecified routing table
			RT_TABLE_DEFAULT  = 0xFD  # The default table
			RT_TABLE_MAIN     = 0xFE  # The main table
			RT_TABLE_LOCAL    = 0xFF  # The local table

		class Protocol (object):
			RTPROT_UNSPEC     = 0x00  # Identifies what/who added the route
			RTPROT_REDIRECT   = 0x01  # By an ICMP redirect
			RTPROT_KERNEL     = 0x02  # By the kernel
			RTPROT_BOOT       = 0x03  # During bootup
			RTPROT_STATIC     = 0x04  # By the administrator
			RTPROT_GATED      = 0x08  # GateD
			RTPROT_RA         = 0x09  # RDISC/ND router advertissements
			RTPROT_MRT        = 0x0A  # Merit MRT
			RTPROT_ZEBRA      = 0x0B  # ZEBRA
			RTPROT_BIRD       = 0x0C  # BIRD
			RTPROT_DNROUTED   = 0x0D  # DECnet routing daemon
			RTPROT_XORP       = 0x0E  # XORP
			RTPROT_NTK        = 0x0F  # Netsukuku
			RTPROT_DHCP       = 0x10  # DHCP client
			# Self allocating ourself a kernel table :p
			# http://lxr.free-electrons.com/source/include/uapi/linux/rtnetlink.h#L237
			# YES WE CAN !
			RTPROT_EXABGP     = 0x11  # Exa Networks ExaBGP

		class Scope (object):
			RT_SCOPE_UNIVERSE = 0x00  # Global route
			RT_SCOPE_SITE     = 0xC8  # Interior route in the local autonomous system
			RT_SCOPE_LINK     = 0xFD  # Route on this link
			RT_SCOPE_HOST     = 0xFE  # Route on the local host
			RT_SCOPE_NOWHERE  = 0xFF  # Destination does not exist

		class Type (object):
			RTN_UNSPEC        = 0x00  # Unknown route.
			RTN_UNICAST       = 0x01  # A gateway or direct route.
			RTN_LOCAL         = 0x02  # A local interface route.
			RTN_BROADCAST     = 0x03  # A local broadcast route (sent as a broadcast).
			RTN_ANYCAST       = 0x04  # An anycast route.
			RTN_MULTICAST     = 0x05  # A multicast route.
			RTN_BLACKHOLE     = 0x06  # A silent packet dropping route.
			RTN_UNREACHABLE   = 0x07  # An unreachable destination.  Packets dropped and host unreachable ICMPs are sent to the originator.
			RTN_PROHIBIT      = 0x08  # A packet rejection route.  Packets are dropped and communication prohibited ICMPs are sent to the originator.
			RTN_THROW         = 0x09  # When used with policy routing, continue routing lookup in another table.  Under normal routing, packets are dropped and net unreachable ICMPs are sent to the originator.
			RTN_NAT           = 0x0A  # A network address translation rule.
			RTN_XRESOLVE      = 0x0B  # Refer to an external resolver (not implemented).

		class Flag (object):
			RTM_F_NOTIFY      = 0x100  # If the route changes, notify the user
			RTM_F_CLONED      = 0x200  # Route is cloned from another route
			RTM_F_EQUALIZE    = 0x400  # Allow randomization of next hop path in multi-path routing (currently not implemented)
			RTM_F_PREFIX      = 0x800  # Prefix Address

		class Attribute (object):
			RTA_UNSPEC        = 0x00  # Ignored.
			RTA_DST           = 0x01  # Protocol address for route destination address.
			RTA_SRC           = 0x02  # Protocol address for route source address.
			RTA_IIF           = 0x03  # Input interface index.
			RTA_OIF           = 0x04  # Output interface index.
			RTA_GATEWAY       = 0x05  # Protocol address for the gateway of the route
			RTA_PRIORITY      = 0x06  # Priority of route.
			RTA_PREFSRC       = 0x07  # Preferred source address in cases where more than one source address could be used.
			RTA_METRICS       = 0x08  # Route metrics attributed to route and associated protocols (e.g., RTT, initial TCP window, etc.).
			RTA_MULTIPATH     = 0x09  # Multipath route next hop's attributes.
			# RTA_PROTOINFO   = 0x0A  # Firewall based policy routing attribute.
			RTA_FLOW          = 0x0B  # Route realm.
			RTA_CACHEINFO     = 0x0C  # Cached route information.
			# RTA_SESSION     = 0x0D
			# RTA_MP_ALGO     = 0x0E
			RTA_TABLE         = 0x0F

	@classmethod
	def getRoutes (cls):
		return cls.extract(Network.Command.RTM_GETROUTE)

	@classmethod
	def newRoute (cls):
		network_flags  = NetLinkMessage.Flags.NLM_F_REQUEST
		network_flags |= NetLinkMessage.Flags.NLM_F_ACK
		network_flags |= NetLinkMessage.Flags.NLM_F_CREATE
#		network_flags |= NetLinkMessage.Flags.NLM_F_EXCL

		family = socket.AF_INET

		prefix = '\x0a\1\0\0'
		prefix_len = 24
		gateway  = '\x0a\0\0\1'

		attributes = {
			Network.Type.Attribute.RTA_DST: prefix,
			Network.Type.Attribute.RTA_GATEWAY: gateway,
		}

		# format = namedtuple('Address', 'family prefixlen flags scope index attributes')
		# format = namedtuple('Neighbor', 'family src_len dst_len tos table proto scope type flags attributes')

		neighbor = cls.format(
			family,
			0,     # src_len
			prefix_len,    # dst_len ( only /32 atm)
			0,     # tos
			Network.Type.Table.RT_TABLE_MAIN,
			Network.Type.Protocol.RTPROT_EXABGP,
			Network.Type.Scope.RT_SCOPE_UNIVERSE,
			Network.Type.Type.RTN_UNICAST,
			Network.Type.Flag.RTM_F_PREFIX,  # this may be wrong
			attributes
		)

		return cls.extract(Network.Command.RTM_NEWROUTE,network_flags,family,neighbor)

	@classmethod
	def delRoute (cls):
		return cls.extract(Network.Command.RTM_DELROUTE)



# encoding: utf-8
"""
reader.py

Created by Thomas Mangin on 2013-03-15.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""


class InvalidFormat (Exception):

	"""Raised when the configuration can not be parsed"""

	pass


class Format (object):

	r"""Text formating class

	Store class used to convert the configuration format to json
	Every configuration file must start with the line "#syntax: <format>\\n"
	where <format> is one of the class defined within format (simplejson or json)
	"""

	class exabgp (object):

		"""raw reader without any modification"""

		@staticmethod
		def skip (current):
			return False

		@staticmethod
		def read (last, current):
			return current


class Reader (object):

	"""File like object

	providing a read() method which will convert the configuration
	in JSON following the format information set at
	the start of the file with the "#syntax: <format>"
	"""

	def __init__ (self, fname):
		self.file = open(fname,'rb')
		self.last = ''      # the last line we read from the file
		self.formated = ''  # the formated data we have already converted

		name = ''.join(self.file.readline().split())
		if not name.startswith('#syntax:'):
			name = '#syntax:exabgp'
			self.file.close()
			self.file = open(fname,'rb')

		klass = getattr(Format,name[8:],None)
		if not klass:
			raise InvalidFormat('unknown configuration format')

		self.format = klass.read
		self.skip = klass.skip

	def __del__ (self):
		if self.file:
			self.file.close()
			self.file = None

	def __enter__ (self):
		return self

	def __exit__ (self, dtype, value, tb):
		if self.file:
			self.file.close()
			self.file = None

	def read (self, number=0):
		# we already done the work, just return the small chunks
		if number and len(self.formated) >= number:
			returned, self.formated = self.formated[:number], self.formated[number:]
			return returned

		data = bytearray()

		try:
			# restore / init the last line seen
			last = self.last

			# reading up to number bytes or until EOF which will raise StopIteration
			while not number or len(data) < number:
				new = self.file.next()
				if self.skip(new):
					continue
				data += self.format(last,new)
				last = new

			# save the last line seen for the next call
			self.last = last

			if number:
				complete = self.formated + bytes(data)
				returned, self.formated = complete[:number], complete[number:]
				return returned

			return bytes(data)
		except StopIteration:
			# we can come here twice : on EOF and again
			# to empty self.formated when its len becomes smaller than number
			complete = self.formated + bytes(data)
			if number:
				returned, self.formated = complete[:number], complete[number:]
				return returned
			else:
				self.formated = ''
				return complete

	def readline (self, limit=-1):
		returned = bytearray()
		while limit < 0 or len(returned) < limit:
			byte = self.read(1)
			if not byte:
				break
			returned += byte
			if returned.endswith(b'\n'):
				break
		return bytes(returned)

	def __iter__ (self):
		if not self.file:
			raise ValueError("I/O operation on closed file.")
		return self

	def next (self):
		line = self.readline()
		if not line:
			raise StopIteration
		return line

	__next__ = next



# encoding: utf-8
"""
routerid.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.family import AFI
from exabgp.protocol.ip import IPv4

# ===================================================================== RouterID
#


class RouterID (IPv4):
	def __init__ (self, ip, packed=None):
		if IPv4.toafi(ip) != AFI.ipv4:
			raise ValueError('wrong address family')
		IPv4.__init__(self,ip,packed)

	@classmethod
	def unpack (cls, data):  # pylint: disable=W0221
		return cls('.'.join(str(ord(_)) for _ in data),data)



# encoding: utf-8
"""
encapsulation.py

Created by Thomas Mangin on 2014-06-20.
Copyright (c) 2014-2015 Orange. All rights reserved.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack

from exabgp.bgp.message.update.attribute.community.extended import ExtendedCommunity

# ============================================================ Layer2Information
# RFC 4761


class L2Info (ExtendedCommunity):
	COMMUNITY_TYPE = 0x80
	COMMUNITY_SUBTYPE = 0x0A

	__slots__ = ['encaps','control','mtu','reserved']

	def __init__ (self, encaps, control, mtu, reserved, community=None):
		self.encaps = encaps
		self.control = control
		self.mtu = mtu
		self.reserved = reserved
		# reserved is called preference in draft-ietf-l2vpn-vpls-multihoming-07
		ExtendedCommunity.__init__(
			self,
			community if community is not None else pack(
				"!2sBBHH",
				self._packedTypeSubtype(),
				encaps,control,
				mtu,reserved
			)
		)

	def __str__ (self):
		return "l2info:%s:%s:%s:%s" % (self.encaps,self.control,self.mtu,self.reserved)

	@staticmethod
	def unpack (data):
		encaps,control,mtu,reserved = unpack('!BBHH',data[2:8])
		return L2Info(encaps,control,mtu,reserved,data[:8])



# encoding: utf-8
"""
fragment.py

Created by Thomas Mangin on 2010-02-04.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.enum import Enum
# =================================================================== Fragment

# Uses bitmask operand format defined above.
#   0   1   2   3   4   5   6   7
# +---+---+---+---+---+---+---+---+
# |   Reserved    |LF |FF |IsF|DF |
# +---+---+---+---+---+---+---+---+
#
# Bitmask values:
# +  Bit 7 - Don't fragment (DF)
# +  Bit 6 - Is a fragment (IsF)
# +  Bit 5 - First fragment (FF)
# +  Bit 4 - Last fragment (LF)

class Fragment (Enum):
	NOT      = 0x00
	DONT     = 0x01
	IS       = 0x02
	FIRST    = 0x04
	LAST     = 0x08
	# reserved = 0xF0


Fragment.UNKNOWN = 'unknown fragment value %s'

Fragment.VALUE = {
	'not-a-fragment':   Fragment(Fragment.NOT),
	'dont-fragment':  Fragment(Fragment.DONT),
	'is-fragment':    Fragment(Fragment.IS),
	'first-fragment': Fragment(Fragment.FIRST),
	'last-fragment':  Fragment(Fragment.LAST),
}

Fragment.NAME = {
	Fragment(Fragment.NOT):   'not-a-fragment',
	Fragment(Fragment.DONT):  'dont-fragment',
	Fragment(Fragment.IS):    'is-fragment',
	Fragment(Fragment.FIRST): 'first-fragment',
	Fragment(Fragment.LAST):  'last-fragment',
}



# encoding: utf-8
"""
family.py

Created by Thomas Mangin on 2014-06-22.
Copyright (c) 2014-2015 Exa Networks. All rights reserved.
"""

from exabgp.configuration.engine.raised import Raised
from exabgp.configuration.engine.registry import Entry
from exabgp.configuration.engine.parser import boolean

from exabgp.bgp.message.open.capability.capability import Capability


# =================================================================== Capability
#

class SectionCapability (Entry):
	syntax = \
		'capability {\n' \
		'   asn4 enable|disable;                         # default enabled\n' \
		'   aigp enable|disable;                         # default disabled\n' \
		'   operational enable|disable;                  # default disabled\n' \
		'   multi-session enable|disable;                # default disabled\n' \
		'   route-refresh enable|disable;                # default disabled\n' \
		'   graceful-restart <time in second>;           # default disabled\n' \
		'   add-path disable|send|receive|send/receive;  # default disabled\n' \
		'}\n'

	def __init__ (self):
		self.content = dict()

	def enter (self, tokeniser):
		token = tokeniser()
		if token != '{':
			raise Raised(self.syntax)
		self.content = dict()

	def exit (self, tokeniser):
		# no verification to do
		pass

	def asn4 (self, tokeniser):
		self.content[Capability.CODE.FOUR_BYTES_ASN] = boolean(tokeniser,True)
		self._drop_colon(tokeniser)

	def aigp (self, tokeniser):
		self.content[Capability.CODE.AIGP] = boolean(tokeniser,False)
		self._drop_colon(tokeniser)

	def addpath (self, tokeniser):
		ap = tokeniser()
		if ap not in ('receive','send','send/receive','disable','disabled'):
			raise Raised("")

		self.content[Capability.CODE.ADD_PATH] = 0
		if ap.endswith('receive'):
			self.content[Capability.CODE.ADD_PATH] += 1
		if ap.startswith('send'):
			self.content[Capability.CODE.ADD_PATH] += 2

		self._drop_colon(tokeniser)

	def operational (self, tokeniser):
		self.content[Capability.CODE.OPERATIONAL] = boolean(tokeniser,False)
		self._drop_colon(tokeniser)

	def refresh (self, tokeniser):
		self.content[Capability.CODE.ROUTE_REFRESH] = boolean(tokeniser,False)
		self._drop_colon(tokeniser)

	def multisession (self, tokeniser):
		self.content[Capability.CODE.MULTISESSION] = boolean(tokeniser,False)
		self._drop_colon(tokeniser)

	def graceful (self, tokeniser):
		token = tokeniser()
		if not token.isdigit():
			raise Raised("")

		duration = int(token)
		if duration < 0:
			raise Raised("")
		if duration > pow(2,16):
			raise Raised("")

		self.content[Capability.CODE.GRACEFUL_RESTART] = duration
		self._drop_colon(tokeniser)

	def _check_duplicate (self, key):
		if key in self.content:
			raise Raised("")

	@classmethod
	def register (cls, registry, location):
		registry.register_class(cls)

		registry.register_hook(cls,'enter',location,'enter')
		registry.register_hook(cls,'exit',location,'exit')

		registry.register_hook(cls,'action',location+['asn4'],'asn4')
		registry.register_hook(cls,'action',location+['aigp'],'aigp')
		registry.register_hook(cls,'action',location+['add-path'],'addpath')
		registry.register_hook(cls,'action',location+['operational'],'operational')
		registry.register_hook(cls,'action',location+['route-refresh'],'refresh')
		registry.register_hook(cls,'action',location+['multi-session'],'multisession')
		registry.register_hook(cls,'action',location+['graceful-restart'],'graceful')



# encoding: utf-8
"""
address.py

Created by Thomas Mangin on 2010-01-19.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from struct import pack
from struct import unpack


# =================================================================== AFI

# http://www.iana.org/assignments/address-family-numbers/
class AFI (int):
	undefined = 0x00  # internal
	ipv4      = 0x01
	ipv6      = 0x02
	l2vpn     = 0x19

	Family = {
		ipv4:  0x02,  # socket.AF_INET,
		ipv6:  0x30,  # socket.AF_INET6,
		l2vpn: 0x02,  # l2vpn info over ipv4 session
	}

	names = {
		'ipv4':  ipv4,
		'ipv6':  ipv6,
		'l2vpn': l2vpn,
	}

	def __str__ (self):
		if self == 0x01:
			return "ipv4"
		if self == 0x02:
			return "ipv6"
		if self == 0x19:
			return "l2vpn"
		return "unknown afi %d" % self

	def __repr__ (self):
		return str(self)

	def name (self):
		if self == 0x01:
			return "inet4"
		if self == 0x02:
			return "inet6"
		if self == 0x19:
			return "l2vpn"
		return "unknown afi"

	def pack (self):
		return pack('!H',self)

	@staticmethod
	def unpack (data):
		return AFI(unpack('!H',data)[0])

	@staticmethod
	def value (name):
		if name == "ipv4":
			return AFI.ipv4
		if name == "ipv6":
			return AFI.ipv6
		return None

	@staticmethod
	def implemented_safi (afi):
		if afi == 'ipv4':
			return ['unicast','multicast','nlri-mpls','mpls-vpn','flow','flow-vpn']
		if afi == 'ipv6':
			return ['unicast','mpls-vpn','flow','flow-vpn']
		if afi == 'l2vpn':
			return ['vpls']
		return []

	@classmethod
	def fromString (cls, string):
		return cls.names.get(string,cls.undefined)


# =================================================================== SAFI

# http://www.iana.org/assignments/safi-namespace
class SAFI (int):
	undefined = 0               # internal
	unicast = 1                 # [RFC4760]
	multicast = 2               # [RFC4760]
	# deprecated = 3            # [RFC4760]
	nlri_mpls = 4               # [RFC3107]
	# mcast_vpn = 5             # [draft-ietf-l3vpn-2547bis-mcast-bgp] (TEMPORARY - Expires 2008-06-19)
	# pseudowire = 6            # [draft-ietf-pwe3-dynamic-ms-pw] (TEMPORARY - Expires 2008-08-23) Dynamic Placement of Multi-Segment Pseudowires
	# encapsulation = 7         # [RFC5512]

	# tunel = 64                # [Nalawade]
	vpls = 65                   # [RFC4761]
	# bgp_mdt = 66              # [Nalawade]
	# bgp_4over6 = 67           # [Cui]
	# bgp_6over4 = 67           # [Cui]
	# vpn_adi = 69              # [RFC-ietf-l1vpn-bgp-auto-discovery-05.txt]

	evpn = 70                   # [draft-ietf-l2vpn-evpn]
	mpls_vpn = 128              # [RFC4364]
	# mcast_bgp_mpls_vpn = 129  # [RFC2547]
	# rt = 132                  # [RFC4684]
	rtc = 132                   # [RFC4684]
	flow_ip = 133               # [RFC5575]
	flow_vpn = 134              # [RFC5575]

	# vpn_ad = 140              # [draft-ietf-l3vpn-bgpvpn-auto]

	# private = [_ for _ in range(241,254)]   # [RFC4760]
	# unassigned = [_ for _ in range(8,64)] + [_ for _ in range(70,128)]
	# reverved = [0,3] + [130,131] + [_ for _ in range(135,140)] + [_ for _ in range(141,241)] + [255,]    # [RFC4760]

	names = {
		'unicast':   unicast,
		'multicast': multicast,
		'nlri-mpls': nlri_mpls,
		'vpls':      vpls,
		'evpn':      evpn,
		'mpls-vpn':  mpls_vpn,
		'rtc':       rtc,
		'flow':      flow_ip,
		'flow-vpn':  flow_vpn,
	}

	def name (self):
		if self == 0x01:
			return "unicast"
		if self == 0x02:
			return "multicast"
		if self == 0x04:
			return "nlri-mpls"
		if self == 0x46:
			return "evpn"
		if self == 0x80:
			return "mpls-vpn"
		if self == 0x84:
			return "rtc"
		if self == 0x85:
			return "flow"
		if self == 0x86:
			return "flow-vpn"
		if self == 0x41:
			return "vpls"
		return "unknown safi %d" % self

	def __str__ (self):
		return self.name()

	def __repr__ (self):
		return str(self)

	def pack (self):
		return chr(self)

	@staticmethod
	def unpack (data):
		return SAFI(ord(data))

	def has_label (self):
		return self in (self.nlri_mpls,self.mpls_vpn)

	def has_rd (self):
		return self in (self.mpls_vpn,)  # technically self.flow_vpn and self.vpls has an RD but it is not an NLRI

	@staticmethod
	def value (name):
		if name == "unicast":
			return 0x01
		if name == "multicast":
			return 0x02
		if name == "nlri-mpls":
			return 0x04
		if name == "mpls-vpn":
			return 0x80
		if name == "flow":
			return 0x85
		if name == "flow-vpn":
			return 0x86
		if name == "vpls":
			return 0x41
		return None

	@classmethod
	def fromString (cls, string):
		return cls.names.get(string,cls.undefined)


def known_families ():
	# it can not be a generator
	families = [
		(AFI(AFI.ipv4), SAFI(SAFI.unicast)),
		(AFI(AFI.ipv4), SAFI(SAFI.multicast)),
		(AFI(AFI.ipv4), SAFI(SAFI.nlri_mpls)),
		(AFI(AFI.ipv4), SAFI(SAFI.mpls_vpn)),
		(AFI(AFI.ipv4), SAFI(SAFI.flow_ip)),
		(AFI(AFI.ipv4), SAFI(SAFI.flow_vpn)),
		(AFI(AFI.ipv6), SAFI(SAFI.unicast)),
		(AFI(AFI.ipv6), SAFI(SAFI.nlri_mpls)),
		(AFI(AFI.ipv6), SAFI(SAFI.mpls_vpn)),
		(AFI(AFI.ipv6), SAFI(SAFI.flow_ip)),
		(AFI(AFI.ipv6), SAFI(SAFI.flow_vpn)),
		(AFI(AFI.l2vpn), SAFI(SAFI.vpls))
	]
	return families


class Family (object):
	def __init__ (self, afi, safi):
		self.afi = AFI(afi)
		self.safi = SAFI(safi)

	def extensive (self):
		return 'afi %s safi %s' % (self.afi,self.safi)

	def __str__ (self):
		return 'family %s %s' % (self.afi,self.safi)



# encoding: utf-8
"""
flow.py

Created by Thomas Mangin on 2010-01-14.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

# Do not use __slots__ here, we never create enough of them to be worth it
# And it really break complex inheritance

from struct import pack
from struct import unpack

from exabgp.protocol.ip import IP
from exabgp.protocol.ip import NoIP
from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI
from exabgp.bgp.message import OUT
from exabgp.bgp.message.notification import Notify
from exabgp.bgp.message.update.nlri.cidr import CIDR

from exabgp.protocol import Protocol
from exabgp.protocol.ip.icmp import ICMPType
from exabgp.protocol.ip.icmp import ICMPCode
from exabgp.protocol.ip.fragment import Fragment
from exabgp.protocol.ip.tcp.flag import TCPFlag

from exabgp.bgp.message.update.nlri.nlri import NLRI
from exabgp.bgp.message.update.nlri.qualifier.rd import RouteDistinguisher


# =================================================================== Flow Components

class IComponent (object):
	# all have ID
	# should have an interface for serialisation and put it here
	pass


class CommonOperator (object):
	# power (2,x) is the same as 1 << x which is what the RFC say the len is
	power = {0:1, 1:2, 2:4, 3:8,}
	rewop = {1:0, 2:1, 4:2, 8:3,}
	len_position = 0x30

	EOL       = 0x80  # 0b10000000
	AND       = 0x40  # 0b01000000
	LEN       = 0x30  # 0b00110000
	NOP       = 0x00

	OPERATOR  = 0xFF ^ (EOL | LEN)

	@staticmethod
	def eol (data):
		return data & CommonOperator.EOL

	@staticmethod
	def operator (data):
		return data & CommonOperator.OPERATOR

	@staticmethod
	def length (data):
		return 1 << ((data & CommonOperator.LEN) >> 4)


class NumericOperator (CommonOperator):
	# reserved= 0x08  # 0b00001000
	LT        = 0x04  # 0b00000100
	GT        = 0x02  # 0b00000010
	EQ        = 0x01  # 0b00000001


class BinaryOperator (CommonOperator):
	# reserved= 0x0C  # 0b00001100
	NOT       = 0x02  # 0b00000010
	MATCH     = 0x01  # 0b00000001


def _len_to_bit (value):
	return NumericOperator.rewop[value] << 4


def _bit_to_len (value):
	return NumericOperator.power[(value & CommonOperator.len_position) >> 4]


def _number (string):
	value = 0
	for c in string:
		value = (value << 8) + ord(c)
	return value

# def short (value):
# 	return (ord(value[0]) << 8) + ord(value[1])

# Interface ..................


class IPv4 (object):
	afi = AFI.ipv4


class IPv6 (object):
	afi = AFI.ipv6


class IPrefix (object):
	pass


# Prococol

class IPrefix4 (IPrefix,IComponent,IPv4):
	# Must be defined in subclasses
	CODE = -1
	NAME = ''

	# not used, just present for simplying the nlri generation
	operations = 0x0

	def __init__ (self, raw, netmask):
		self.nlri = CIDR(raw,netmask)

	def pack (self):
		raw = self.nlri.pack()
		# ID is defined in subclasses
		return "%s%s" % (chr(self.ID),raw)  # pylint: disable=E1101

	def __str__ (self):
		return str(self.nlri)


class IPrefix6 (IPrefix,IComponent,IPv6):
	# Must be defined in subclasses
	CODE = -1
	NAME = ''

	# not used, just present for simplying the nlri generation
	operations = 0x0

	def __init__ (self, raw, netmask, offset):
		self.nlri = CIDR(raw,netmask)
		self.offset = offset

	def pack (self):
		raw = self.nlri.packed_ip()
		# ID is defined in subclasses
		return "%s%s%s%s" % (chr(self.ID),chr(self.nlri.mask),chr(self.offset),raw)  # pylint: disable=E1101

	def __str__ (self):
		return "%s/%s" % (self.nlri,self.offset)


class IOperation (IComponent):
	# need to implement encode which encode the value of the operator

	def __init__ (self, operations, value):
		self.operations = operations
		self.value = value
		self.first = None  # handled by pack/str

	def pack (self):
		l,v = self.encode(self.value)
		op = self.operations | _len_to_bit(l)
		return "%s%s" % (chr(op),v)

	def encode (self, value):
		raise NotImplementedError('this method must be implemented by subclasses')

	def decode (self, value):
		raise NotImplementedError('this method must be implemented by subclasses')


# class IOperationIPv4 (IOperation):
# 	def encode (self, value):
# 		return 4, socket.pton(socket.AF_INET,value)

class IOperationByte (IOperation):
	def encode (self, value):
		return 1,chr(value)

	def decode (self, bgp):
		return ord(bgp[0]),bgp[1:]


class IOperationByteShort (IOperation):
	def encode (self, value):
		if value < (1 << 8):
			return 1,chr(value)
		return 2,pack('!H',value)

	def decode (self, bgp):
		return unpack('!H',bgp[:2])[0],bgp[2:]


class IOperationByteShortLong (IOperation):
	def encode (self, value):
		if value < (1 << 8):
			return 1,chr(value)
		if value < (1 << 16):
			return 2,pack('!H',value)
		return 4,pack('!L',value)

	# XXX: buggy ?? as it assumes 4 bytes but may be less
	def decode (self, bgp):
		return unpack('!L',bgp[:4])[0],bgp[4:]


# String representation for Numeric and Binary Tests

class NumericString (object):
	operations = None
	value = None

	_string = {
		NumericOperator.LT: '<',
		NumericOperator.GT: '>',
		NumericOperator.EQ: '=',
		NumericOperator.LT | NumericOperator.EQ: '<=',
		NumericOperator.GT | NumericOperator.EQ: '>=',

		NumericOperator.AND | NumericOperator.LT: '&<',
		NumericOperator.AND | NumericOperator.GT: '&>',
		NumericOperator.AND | NumericOperator.EQ: '&=',
		NumericOperator.AND | NumericOperator.LT | NumericOperator.EQ: '&<=',
		NumericOperator.AND | NumericOperator.GT | NumericOperator.EQ: '&>=',
	}

	def __str__ (self):
		return "%s%s" % (self._string[self.operations & (CommonOperator.EOL ^ 0xFF)], self.value)


class BinaryString (object):
	operations = None
	value = None

	_string = {
		BinaryOperator.NOT:   '!',
		BinaryOperator.MATCH: '=',
		BinaryOperator.AND | BinaryOperator.NOT:   '&!',
		BinaryOperator.AND | BinaryOperator.MATCH: '&=',
	}

	def __str__ (self):
		return "%s%s" % (self._string[self.operations & (CommonOperator.EOL ^ 0xFF)], self.value)


# Components ..............................

def converter (function, klass=None):
	def _integer (value):
		if klass is None:
			return function(value)
		try:
			return klass(value)
		except ValueError:
			return function(value)
	return _integer


def decoder (function, klass=int):
	def _inner (value):
		return klass(function(value))
	return _inner


def PacketLength (data):
	_str_bad_length = "cloudflare already found that invalid max-packet length for for you .."
	number = int(data)
	if number > 0xFFFF:
		raise ValueError(_str_bad_length)
	return number


def PortValue (data):
	_str_bad_port = "you tried to set an invalid port number .."
	number = int(data)
	if number < 0 or number > 0xFFFF:
		raise ValueError(_str_bad_port)
	return number


def DSCPValue (data):
	_str_bad_dscp = "you tried to filter a flow using an invalid dscp for a component .."
	number = int(data)
	if number < 0 or number > 0xFFFF:
		raise ValueError(_str_bad_dscp)
	return number


def ClassValue (data):
	_str_bad_class = "you tried to filter a flow using an invalid traffic class for a component .."
	number = int(data)
	if number < 0 or number > 0xFFFF:
		raise ValueError(_str_bad_class)
	return number


def LabelValue (data):
	_str_bad_label = "you tried to filter a flow using an invalid traffic label for a component .."
	number = int(data)
	if number < 0 or number > 0xFFFFF:  # 20 bits 5 bytes
		raise ValueError(_str_bad_label)
	return number


# Protocol Shared

class FlowDestination (object):
	ID = 0x01
	NAME = 'destination'


class FlowSource (object):
	ID = 0x02
	NAME = 'source'


# Prefix
class Flow4Destination (IPrefix4,FlowDestination):
	NAME = 'destination'


# Prefix
class Flow4Source (IPrefix4,FlowSource):
	NAME = 'source'


# Prefix
class Flow6Destination (IPrefix6,FlowDestination):
	NAME = 'destination'


# Prefix
class Flow6Source (IPrefix6,FlowSource):
	NAME = 'source'


class FlowIPProtocol (IOperationByte,NumericString,IPv4):
	ID  = 0x03
	NAME = 'protocol'
	converter = staticmethod(converter(Protocol.Name,Protocol))
	decoder = staticmethod(decoder(ord,Protocol))


class FlowNextHeader (IOperationByte,NumericString,IPv6):
	ID  = 0x03
	NAME = 'next-header'
	converter = staticmethod(converter(Protocol.Name,Protocol))
	decoder = staticmethod(decoder(ord,Protocol))


class FlowAnyPort (IOperationByteShort,NumericString,IPv4,IPv6):
	ID  = 0x04
	NAME = 'port'
	converter = staticmethod(converter(PortValue))
	decoder = staticmethod(_number)


class FlowDestinationPort (IOperationByteShort,NumericString,IPv4,IPv6):
	ID  = 0x05
	NAME = 'destination-port'
	converter = staticmethod(converter(PortValue))
	decoder = staticmethod(_number)


class FlowSourcePort (IOperationByteShort,NumericString,IPv4,IPv6):
	ID  = 0x06
	NAME = 'source-port'
	converter = staticmethod(converter(PortValue))
	decoder = staticmethod(_number)


class FlowICMPType (IOperationByte,NumericString,IPv4,IPv6):
	ID = 0x07
	NAME = 'icmp-type'
	converter = staticmethod(converter(ICMPType.Name))
	decoder = staticmethod(decoder(_number,ICMPType))


class FlowICMPCode (IOperationByte,NumericString,IPv4,IPv6):
	ID = 0x08
	NAME = 'icmp-code'
	converter = staticmethod(converter(ICMPCode.Name))
	decoder = staticmethod(decoder(_number,ICMPCode))


class FlowTCPFlag (IOperationByte,BinaryString,IPv4,IPv6):
	ID = 0x09
	NAME = 'tcp-flags'
	converter = staticmethod(converter(TCPFlag.Name))
	decoder = staticmethod(decoder(ord,TCPFlag))


class FlowPacketLength (IOperationByteShort,NumericString,IPv4,IPv6):
	ID = 0x0A
	NAME = 'packet-length'
	converter = staticmethod(converter(PacketLength))
	decoder = staticmethod(_number)


# RFC2474
class FlowDSCP (IOperationByteShort,NumericString,IPv4):
	ID = 0x0B
	NAME = 'dscp'
	converter = staticmethod(converter(DSCPValue))
	decoder = staticmethod(_number)


# RFC2460
class FlowTrafficClass (IOperationByte,NumericString,IPv6):
	ID = 0x0B
	NAME = 'traffic-class'
	converter = staticmethod(converter(ClassValue))
	decoder = staticmethod(_number)


# BinaryOperator
class FlowFragment (IOperationByteShort,BinaryString,IPv4):
	ID = 0x0C
	NAME = 'fragment'
	converter = staticmethod(converter(Fragment.Name))
	decoder = staticmethod(decoder(ord,Fragment))


# draft-raszuk-idr-flow-spec-v6-01
class FlowFlowLabel (IOperationByteShortLong,NumericString,IPv6):
	ID = 0x0D
	NAME = 'flow-label'
	converter = staticmethod(converter(LabelValue))
	decoder = staticmethod(_number)


# ..........................................................

decode = {AFI.ipv4: {}, AFI.ipv6: {}}
factory = {AFI.ipv4: {}, AFI.ipv6: {}}

for content in dir():
	kls = globals().get(content,None)
	if not isinstance(kls,type(IComponent)):
		continue
	if not issubclass(kls,IComponent):
		continue

	_ID = getattr(kls,'ID',None)
	if not _ID:
		continue

	_afis = []
	if issubclass(kls,IPv4):
		_afis.append(AFI.ipv4)
	if issubclass(kls,IPv6):
		_afis.append(AFI.ipv6)

	for _afi in _afis:
		factory[_afi][_ID] = kls
		name = getattr(kls,'NAME')

		if issubclass(kls, IOperation):
			if issubclass(kls, BinaryString):
				decode[_afi][_ID] = 'binary'
			elif issubclass(kls, NumericString):
				decode[_afi][_ID] = 'numeric'
			else:
				raise RuntimeError('invalid class defined (string)')
		elif issubclass(kls, IPrefix):
			decode[_afi][_ID] = 'prefix'
		else:
			raise RuntimeError('unvalid class defined (type)')


# ..........................................................

def _unique ():
	value = 0
	while True:
		yield value
		value += 1


unique = _unique()


class Flow (NLRI):
	def __init__ (self, afi=AFI.ipv4,safi=SAFI.flow_ip,nexthop=None,rd=None):
		NLRI.__init__(self,afi,safi)
		self.rules = {}
		self.action = OUT.ANNOUNCE
		self.nexthop = IP.unpack(nexthop) if nexthop else NoIP
		self.rd = rd
		self.unique = unique.next()

	def __eq__ (self,other):
		return self.index() == other.index()

	def __ne__ (self, other):
		return not self.__eq__(other)

	def __hash__ (self):
		return hash(self.pack())

	def __len__ (self):
		return len(self.pack())

	def add (self, rule):
		ID = rule.ID
		if ID in (FlowDestination.ID,FlowSource.ID):
			if ID in self.rules:
				return False
			if ID == FlowDestination.ID:
				pair = self.rules.get(FlowSource.ID,[])
			else:
				pair = self.rules.get(FlowDestination.ID,[])
			if pair:
				if rule.afi != pair[0].afi:
					return False
		self.rules.setdefault(ID,[]).append(rule)
		return True

	# The API requires addpath, but it is irrelevant here.
	def pack (self, addpath=None):
		ordered_rules = []
		# the order is a RFC requirement
		for ID in sorted(self.rules.keys()):
			rules = self.rules[ID]
			# for each component get all the operation to do
			# the format use does not prevent two opposing rules meaning that no packet can ever match
			for rule in rules:
				rule.operations &= (CommonOperator.EOL ^ 0xFF)
			rules[-1].operations |= CommonOperator.EOL
			# and add it to the last rule
			if ID not in (FlowDestination.ID,FlowSource.ID):
				ordered_rules.append(chr(ID))
			ordered_rules.append(''.join(rule.pack() for rule in rules))

		components = ''.join(ordered_rules)

		if self.safi == SAFI.flow_vpn:
			components = self.rd.pack() + components

		l = len(components)
		if l < 0xF0:
			data = "%s%s" % (chr(l),components)
		elif l < 0x0FFF:
			data = "%s%s" % (pack('!H',l | 0xF000),components)
		else:
			raise Notify(3,0,"rule too big for NLRI - how to handle this - does this work ?")
			# data = "%s" % chr(0)

		return data

	def extensive (self):
		string = []
		for index in sorted(self.rules):
			rules = self.rules[index]
			s = []
			for idx,rule in enumerate(rules):
				# only add ' ' after the first element
				if idx and not rule.operations & NumericOperator.AND:
					s.append(' ')
				s.append(rule)
			line = ''.join(str(_) for _ in s)
			if len(s) > 1:
				line = '[ %s ]' % line
			string.append(' %s %s' % (rules[0].NAME,line))
		nexthop = ' next-hop %s' % self.nexthop if self.nexthop is not NoIP else ''
		rd = str(self.rd) if self.rd else ''
		return 'flow' + ''.join(string) + rd + nexthop

	def __str__ (self):
		return self.extensive()

	def _json (self):
		string = []
		for index in sorted(self.rules):
			rules = self.rules[index]
			s = []
			for idx,rule in enumerate(rules):
				# only add ' ' after the first element
				if idx and not rule.operations & NumericOperator.AND:
					s.append(', ')
				s.append('"%s"' % rule)
			string.append(' "%s": [ %s ]' % (rules[0].NAME,''.join(str(_) for _ in s).replace('""','')))
		nexthop = ', "next-hop": "%s"' % self.nexthop if self.nexthop is not NoIP else ''
		rd = ', %s' % self.rd.json() if self.rd else ''
		compatibility = ', "string": "%s"' % self.extensive()
		return '{' + ','.join(string) + rd + nexthop + compatibility + ' }'

	def json (self):
		# this is a stop gap so flow route parsing does not crash exabgp
		# delete unique when this is fixed
		return '"flow-%d": %s' % (self.unique,self._json())

	def index (self):
		return self.pack()

	@classmethod
	def unpack (cls, afi, safi, bgp, has_multiple_path, nexthop, action):
		total = len(bgp)
		length,bgp = ord(bgp[0]),bgp[1:]

		if length & 0xF0 == 0xF0:  # bigger than 240
			extra,bgp = ord(bgp[0]),bgp[1:]
			length = ((length & 0x0F) << 16) + extra

		if length > len(bgp):
			raise Notify(3,10,'invalid length at the start of the the flow')

		bgp = bgp[:length]
		nlri = Flow(afi,safi,nexthop)
		nlri.action = action

		if safi == SAFI.flow_vpn:
			nlri.rd = RouteDistinguisher(bgp[:8])
			bgp = bgp[8:]

		seen = []

		while bgp:
			what,bgp = ord(bgp[0]),bgp[1:]

			if what not in decode.get(afi,{}):
				raise Notify(3,10,'unknown flowspec component received for address family %d' % what)

			seen.append(what)
			if sorted(seen) != seen:
				raise Notify(3,10,'components are not sent in the right order %s' % seen)

			decoded = decode[afi][what]
			klass = factory[afi][what]

			if decoded == 'prefix':
				if afi == AFI.ipv4:
					_,rd,_,mask,size,prefix,left = NLRI._nlri(afi,safi,bgp,action,False)
					adding = klass(prefix,mask)
					if not nlri.add(adding):
						raise Notify(3,10,'components are incompatible (two sources, two destinations, mix ipv4/ipv6) %s' % seen)
					# logger.parser(LazyFormat("added flow %s (%s) payload " % (klass.NAME,adding),bgp[:-len(left)]))
					bgp = left
				else:
					byte,bgp = bgp[1],bgp[0]+bgp[2:]
					offset = ord(byte)
					_,rd,_,mask,size,prefix,left = NLRI._nlri(afi,safi,bgp,action,False)
					adding = klass(prefix,mask,offset)
					if not nlri.add(adding):
						raise Notify(3,10,'components are incompatible (two sources, two destinations, mix ipv4/ipv6) %s' % seen)
					# logger.parser(LazyFormat("added flow %s (%s) payload " % (klass.NAME,adding),bgp[:-len(left)]))
					bgp = left
			else:
				end = False
				while not end:
					byte,bgp = ord(bgp[0]),bgp[1:]
					end = CommonOperator.eol(byte)
					operator = CommonOperator.operator(byte)
					length = CommonOperator.length(byte)
					value,bgp = bgp[:length],bgp[length:]
					adding = klass.decoder(value)
					nlri.add(klass(operator,adding))
					# logger.parser(LazyFormat("added flow %s (%s) operator %d len %d payload " % (klass.NAME,adding,byte,length),value))

		return total-len(bgp),nlri



# encoding: utf-8
"""
bgp.py

Created by Thomas Mangin on 2012-07-08.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.protocol.family import AFI
from exabgp.protocol.family import SAFI

from exabgp.protocol.ip import IP
from exabgp.protocol.ip import NoIP
from exabgp.bgp.message.update.nlri.nlri import NLRI
from exabgp.bgp.message.update.nlri.cidr import CIDR
from exabgp.bgp.message.update.nlri.qualifier.labels import Labels
from exabgp.bgp.message.update.nlri.qualifier.rd import RouteDistinguisher
from exabgp.bgp.message.update.nlri.qualifier.path import PathInfo


# ====================================================== Both MPLS and Inet NLRI
# RFC ....

class MPLS (NLRI,CIDR):
	__slots__ = ['labels','rd','nexthop','action']

	def __init__ (self, afi, safi, packed, mask, nexthop, action):
		self.path_info = PathInfo.NOPATH
		self.labels = Labels.NOLABEL
		self.rd = RouteDistinguisher.NORD
		self.nexthop = IP.unpack(nexthop) if nexthop else NoIP
		self.action = action
		NLRI.__init__(self,afi,safi)
		CIDR.__init__(self,packed,mask)

	def index (self):
		return self.pack() + self.nexthop.pack()

	def has_label (self):
		if self.afi == AFI.ipv4 and self.safi in (SAFI.nlri_mpls,SAFI.mpls_vpn):
			return True
		if self.afi == AFI.ipv6 and self.safi == SAFI.mpls_vpn:
			return True
		return False

	def extensive (self):
		nexthop = ' next-hop %s' % self.nexthop if self.nexthop else ''
		return "%s%s%s%s%s" % (self.prefix(),nexthop,str(self.labels),str(self.path_info),str(self.rd))

	def __len__ (self):
		return CIDR.__len__(self) + len(self.labels) + len(self.rd)

	def __str__ (self):
		return self.extensive()

	def __eq__ (self,other):
		return self.index() == other.index()

	def __ne__ (self, other):
		return not self.__eq__(other)

	def __hash__ (self):
		return hash(self.pack())

	def json (self, announced=True):
		label = self.labels.json()
		rdist = self.rd.json()
		pinfo = self.path_info.json()

		r = []
		if announced:
			if label:
				r.append(label)
			if rdist:
				r.append(rdist)
			if pinfo:
				r.append(pinfo)
		return '"%s": { %s }' % (self.prefix(),", ".join(r))

	def pack (self, addpath=None):
		if not self.has_label():
			if addpath:
				return self.path_info.pack() + CIDR.pack(self)
			return CIDR.pack(self)

		length = len(self.labels)*8 + len(self.rd)*8 + self.mask
		if addpath:
			return self.path_info.pack() + chr(length) + self.labels.pack() + self.rd.pack() + CIDR.packed_ip(self)
		return chr(length) + self.labels.pack() + self.rd.pack() + CIDR.packed_ip(self)

	def index (self):
		return self.pack(self.path_info != PathInfo.NOPATH)

	@classmethod
	def unpack (cls, afi, safi, bgp, addpath, nexthop, action):
		labels,rd,path_identifier,mask,size,prefix,left = NLRI._nlri(afi,safi,bgp,action,addpath)

		nlri = cls(afi,safi,prefix,mask,nexthop,action)
		if labels:
			nlri.labels = Labels(labels)
		if rd:
			nlri.rd = RouteDistinguisher(rd)
		if path_identifier:
			nlri.path_info = PathInfo(None,None,path_identifier)

		return len(bgp) - len(left),nlri



# encoding: utf-8
"""
timer.py

Created by Thomas Mangin on 2012-07-21.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

import time

from exabgp.logger import Logger
from exabgp.bgp.message.nop import _NOP
from exabgp.bgp.message.keepalive import KeepAlive
from exabgp.bgp.message.notification import Notify

# ================================================================ ReceiveTimer
# Track the time for keepalive updates


class ReceiveTimer (object):
	def __init__ (self, me, holdtime, code, subcode, message=''):
		self.logger = Logger()
		self.me = me

		self.holdtime = holdtime
		self.last_read = time.time()

		self.code = code
		self.subcode = subcode
		self.message = message

	def check_ka (self, message=_NOP,ignore=_NOP.TYPE):
		if message.TYPE != ignore:
			self.last_read = time.time()
		if self.holdtime:
			left = int(self.last_read  + self.holdtime - time.time())
			self.logger.timers(self.me('Receive Timer %d second(s) left' % left))
			if left <= 0:
				raise Notify(self.code,self.subcode,self.message)
		elif message.TYPE == KeepAlive.TYPE:
			raise Notify(2,6,'Negotiated holdtime was zero, it was invalid to send us a keepalive messages')


class SendTimer (object):
	def __init__ (self, me, holdtime):
		self.logger = Logger()
		self.me = me

		self.keepalive = holdtime.keepalive()
		self.last_sent = int(time.time())

	def need_ka (self):
		if not self.keepalive:
			return False

		now  = int(time.time())
		left = self.last_sent + self.keepalive - now

		if now != self.last_sent:
			self.logger.timers(self.me('Send Timer %d second(s) left' % left))

		if left <= 0:
			self.last_sent = now
			return True
		return False



# encoding: utf-8
"""
nop.py

Created by Thomas Mangin on 2009-11-06.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message import Message

# ========================================================================= NOP
#


class NOP (Message):
	ID = Message.CODE.NOP
	TYPE = chr(Message.CODE.NOP)

	def message (self):
		return self._message(self.data)

	def __str__ (self):
		return "NOP"

	@classmethod
	def unpack_message (cls, data, negotiated):  # pylint: disable=W0613
		return NOP()

_NOP = NOP()



# encoding: utf-8
"""
capability.py

Created by Thomas Mangin on 2012-07-17.
Copyright (c) 2009-2015 Exa Networks. All rights reserved.
"""

from exabgp.bgp.message.notification import Notify


# =================================================================== Capability
#

class Capability (object):

	class CODE (int):
		__slots__ = []

		RESERVED                 = 0x00  # [RFC5492]
		MULTIPROTOCOL            = 0x01  # [RFC2858]
		ROUTE_REFRESH            = 0x02  # [RFC2918]
		OUTBOUND_ROUTE_FILTERING = 0x03  # [RFC5291]
		MULTIPLE_ROUTES          = 0x04  # [RFC3107]
		EXTENDED_NEXT_HOP        = 0x05  # [RFC5549]
		# 6-63      Unassigned
		GRACEFUL_RESTART         = 0x40  # [RFC4724]
		FOUR_BYTES_ASN           = 0x41  # [RFC4893]
		# 66        Deprecated
		DYNAMIC_CAPABILITY       = 0x43  # [Chen]
		MULTISESSION             = 0x44  # [draft-ietf-idr-bgp-multisession]
		ADD_PATH                 = 0x45  # [draft-ietf-idr-add-paths]
		ENHANCED_ROUTE_REFRESH   = 0x46  # [draft-ietf-idr-bgp-enhanced-route-refresh]
		OPERATIONAL              = 0x47  # ExaBGP only ...
		# 70-127    Unassigned
		ROUTE_REFRESH_CISCO      = 0x80  # I Can only find reference to this in the router logs
		# 128-255   Reserved for Private Use [RFC5492]
		MULTISESSION_CISCO       = 0x83  # What Cisco really use for Multisession (yes this is a reserved range in prod !)

		EXTENDED_MESSAGE         = -1    # No yet defined by draft http://tools.ietf.org/html/draft-ietf-idr-extended-messages-02.txt

		unassigned = range(70,128)
		reserved = range(128,256)

		# Internal
		AIGP = 0xFF00

		names = {
			RESERVED:                  'reserved',
			MULTIPROTOCOL:             'multiprotocol',
			ROUTE_REFRESH:             'route-refresh',
			OUTBOUND_ROUTE_FILTERING:  'outbound-route-filtering',
			MULTIPLE_ROUTES:           'multiple-routes',
			EXTENDED_NEXT_HOP:         'extended-next-hop',

			GRACEFUL_RESTART:          'graceful-restart',
			FOUR_BYTES_ASN:            'asn4',

			DYNAMIC_CAPABILITY:        'dynamic-capability',
			MULTISESSION:              'multi-session',
			ADD_PATH:                  'add-path',
			ENHANCED_ROUTE_REFRESH:    'enhanced-route-refresh',
			OPERATIONAL:               'operational',

			ROUTE_REFRESH_CISCO:       'cisco-route-refresh',
			MULTISESSION_CISCO:        'cisco-multi-sesion',

			AIGP:                      'aigp',
		}

		def __str__ (self):
			name = self.names.get(self,None)
			if name is None:
				if self in Capability.CODE.unassigned:
					return 'unassigned-%s' % hex(self)
				if self in Capability.CODE.reserved:
					return 'reserved-%s' % hex(self)
				return 'capability-%s' % hex(self)
			return name

		def __repr__ (self):
			return str(self)

		@classmethod
		def name (cls, self):
			name = cls.names.get(self,None)
			if name is None:
				if self in Capability.CODE.unassigned:
					return 'unassigned-%s' % hex(self)
				if self in Capability.CODE.reserved:
					return 'reserved-%s' % hex(self)
			return name

	registered_capability = dict()
	_fallback_capability = None

	@staticmethod
	def hex (data):
		return '0x' + ''.join('%02x' % ord(_) for _ in data)

	@classmethod
	def fallback_capability (cls, imp):
		if cls._fallback_capability is not None:
			raise RuntimeError('only one fallback function can be registered')
		cls._fallback_capability = imp

	@staticmethod
	def register_capability (klass, capability=None):
		# ID is defined by all the subclasses - otherwise they do not work :)
		what = klass.ID if capability is None else capability  # pylint: disable=E1101
		if what in klass.registered_capability:
			raise RuntimeError('only one class can be registered per capability')
		klass.registered_capability[what] = klass

	@classmethod
	def klass (cls, what):
		if what in cls.registered_capability:
			kls = cls.registered_capability[what]
			kls.ID = what
			return kls
		if cls._fallback_capability:
			return cls._fallback_capability
		raise Notify (2,4,'can not handle capability %s' % what)

	@classmethod
	def unpack (cls, capability, capabilities, data):
		instance = capabilities.get(capability,Capability.klass(capability)())
		return cls.klass(capability).unpack_capability(instance,data,capability)
